{$I DEBUG}
unit Fonts;

interface

uses VGAGraph, Objects, Memory, RStream;

type
  PFont = ^TFont;
  TFont = object(TObject)
    VSize,
    MaxHSize:	Byte;
    Table:	array[ 0..255 ] of record
				     HSize: 	Byte;
				     Data:	Pointer;
				   end;
    constructor Load( Name: String );
    procedure WriteStr( Origin: TPoint; P: PImage; S: String );
    function CreateStr( S: String ): PSprite;
    procedure MeasureStr( var R: TPoint; S: String );
    destructor Done; virtual;
  end;

implementation

constructor TFont.Load( Name: String );
var
  S:		TRStream;
  I:		Integer;
  OldPos,
  NewPos:	LongInt;
begin
  S.init( Name );
  if S.Status <> stOk then Fail;
  S.Read( VSize, SizeOf(VSize) );
  for I := 0 to 255 do
  begin
    S.Read( Table[I].HSize, SizeOf(Table[I].HSize) );
    S.Read( NewPos, SizeOf(NewPos) );
    if Table[I].HSize > 0 then
    begin
      OldPos := S.GetPos;
      S.Seek( NewPos );
      Table[I].Data := DosAllocMem( VSize*Table[I].HSize );
      if Table[I].Data = nil then
	Halt;
      S.Read( Table[I].Data^, VSize*Table[I].HSize );
      S.Seek( OldPos );
    end;
  end;
  S.Done;
end;

procedure TFont.WriteStr( Origin: TPoint; P: PImage; S: String );
var
  I:		Integer;
  Ofs:		Word;
  P1,
  P2:		Pointer;
  H1,
  H2:		Word;
  VS:		Byte;
begin
  Ofs := Origin.Y*P^.Size.X + Origin.X;
  VS := VSize;
  for I := 1 to Length(S) do
  begin
    with Table[ Byte(S[I]) ] do
    begin
      H1 := HSize;
      if H1 = 0 then Continue;
      P1 := Data;
      H2 := P^.Size.X - HSize;
    end;
    P2 := P^.Data;
    asm
      push ds
      mov bl,VS
      les di,P2
      lds si,P1
      mov di,Ofs
      mov ax,H1
      mov dx,H2
@loop:
      mov cx,ax
      rep movsb
      add di,dx
      dec bl
      jnz @loop
      pop ds
    end;
    Inc( Ofs, H1 );
  end;
end;

function TFont.CreateStr( S: String ): PSprite;
var
  I,
  H:	Integer;
  Dummy,
  Size:	TPoint;
  P:	PSprite;
const
  O:	TPoint = (X:1;Y:1);
begin
  H := 0;
  for I := 1 to Length(S) do
    Inc( H, Table[ Byte(S[I]) ].HSize );
  Size.X := H + 2;
  Size.Y := VSize + 2;
  Dummy.X := 0; Dummy.Y := 0;
  New( P, Init( Dummy, Size, Dummy, nil ) );
  if P <> nil then
    WriteStr( O, P, S );
  CreateStr := P;
end;

procedure TFont.MeasureStr( var R: TPoint; S: String );
var
  I:	Integer;
begin
  R.X := 0;
  for I := 1 to Length(S) do
    Inc( R.X, Table[ Byte(S[I]) ].HSize );
  R.Y := VSize;
end;

destructor TFont.Done;
var
  I:	Integer;
begin
  for I := 0 to 255 do
    if Table[I].HSize <> 0 then
      DosFreeMem( Table[I].Data );
end;

end.