{$I DEBUG}
unit PCX;

interface

uses Objects, RStream;

type
  PRGBPalette = ^TRGBPalette;
  TRGBPalette = array[ 0..255 ] of record R,G,B: Byte; end;

  PCXHeader = record
    Creator,
    Version,
    Encoding,
    Bits:	Byte;
    XMin,
    YMin,
    XMax,
    YMax,
    HRes,
    VRes:	Integer;
    Palette:	array [ 1..48 ] of Byte;
    VMode,
    Planes:	Byte;
    BytesPerLine,
    PaletteInfo,
    SHRes,
    SVRes:	Word;
    Dummy:	array [0..53] of Byte;
  end;

const
  pcxOK			= 0;
  pcxInvalidType	= 1;
  pcxNoFile		= 2;

type
  OldPCXPicture = object
    H:		PCXHeader;
    S:		TRStream;
    Pal:	TRGBPalette;
    Status:	Integer;
    constructor Init( AFileName: String );
    procedure ReadLine( var Buffer );
    function ErrorText: String;
    destructor Done;
  end;

  NewPCXPicture = object
    H:		PCXHeader;
    S:		TBufStream;
    Pal:	TRGBPalette;
    constructor Init( AFileName: String; HSize: Integer );
    procedure WriteLine( var Buffer );
    destructor Done;
  end;

implementation

uses VGAGraph;

type
  PutByteFunc	= procedure( B: Byte );
  GetByteFunc	= function: Byte;
  ByteArr	= array [0..65534] of Byte;
  PByte		= ^ByteArr;

function PackString( var Source; PutByte: PutByteFunc; Size: Integer ): Integer;
var
  SourcePtr:	PByte;
  Count,
  RCount:	Integer;
begin
  Count := 0;
  SourcePtr := @Source;
  repeat
    RCount := 1;
    while (SourcePtr^[Count] = SourcePtr^[Count+RCount]) and (RCount<63) and
	  (Count+RCount < Size ) do
      Inc( RCount );
    if( RCount > 1 ) then
    begin
      PutByte( RCount or $C0 );
      PutByte( SourcePtr^[Count] );
    end
    else
    begin
      if SourcePtr^[Count] >= $C0 then
	PutByte( 1 or $C0 );
      PutByte( SourcePtr^[Count] );
    end;
    Inc( Count, RCount );
  until Count >= Size;
end;

procedure UnpackString( GetByte: GetByteFunc; var Dest; Size: Integer );
var
  DestPtr:	PByte;
  Count:	Integer;
  B:		Byte;
  I:		Integer;
begin
  DestPtr := @Dest;
  Count := 0;
  while Count < Size do
  begin
    B := GetByte;
    if B < $C0 then
    begin
      DestPtr^[Count] := B;
      Inc(Count);
    end
    else
    begin
      DestPtr^[Count] := GetByte;
      for I := 0 to B-$C1 do
	DestPtr^[Count+I] := DestPtr^[Count];
      Inc( Count, I+1 );
    end;
  end;
end;

constructor OldPCXPicture.Init( AFileName: String );
begin
  S.Init( AFileName );
  if S.Status <> stOk then
  begin
    Status := pcxNoFile;
{$IFDEF DEBUG}
    WriteLn( MemF, 'PCX: Error ', ErrorText );
{$ENDIF}
    Exit;
  end;
  S.Read( H, SizeOf(H) );
  if (H.Planes <> 1) or (H.Encoding <> 1) or (H.Bits <> 8 ) then
  begin
    Status := pcxInvalidType;
{$IFDEF DEBUG}
    WriteLn( MemF, 'PCX: Error ', ErrorText );
{$ENDIF}
    Exit;
  end;
  S.Seek( S.GetSize - SizeOf(Pal) );
  S.Read( Pal, SizeOf(Pal) );
  S.Seek( SizeOf(H) );
  Status := pcxOK;
end;

var
  __GetS__:	PStream;

function Get: Byte; far;
var
  B:	Byte;
begin
  __GetS__^.Read( B, 1 );
  Get := B;
end;

procedure OldPCXPicture.ReadLine( var Buffer );
begin
  __GetS__ := @S;
  UnpackString( Get, Buffer, H.BytesPerLine );
end;

function OldPCXPicture.ErrorText: String;
begin
  case Status of
    pcxOK:
      ErrorText := 'No errors';
    pcxNoFile:
      ErrorText := 'Can''t open file';
    pcxInvalidType:
      ErrorText := 'Only grey scale PCXs are supported';
  end;
end;

destructor OldPCXPicture.Done;
begin
  S.Done;
end;

constructor NewPCXPicture.Init( AFileName: String; HSize: Integer );
var
  I:	Integer;
begin
  S.Init( AFileName, stCreate, 4096 );
  if S.Status <> stOk then Fail;
  with H do
  begin
    Creator	:= 10;
    Version	:= 5;
    Encoding	:= 1;
    Bits	:= 8;
    XMin	:= 1;
    YMin	:= 1;
    XMax	:= HSize;
    YMax	:= 0;
    HRes	:= $96;
    VRes	:= $96;
    FillChar( Palette, SizeOf(Palette), 0 );
    VMode	:= 0;
    Planes	:= 1;
    BytesPerLine:= HSize;
    PaletteInfo	:= 1;
    SHRes	:= 0;
    SVRes	:= 0;
  end;
  S.Write( H, SizeOf(H) );
  for I := 0 to 255 do
  begin
    Pal[I].R := I;
    Pal[I].G := I;
    Pal[I].B := I;
  end;
end;

var
  __PutS__:	PStream;

procedure Put( B: Byte ); far;
begin
  __PutS__^.Write( B, 1 );
end;

procedure NewPCXPicture.WriteLine( var Buffer );
begin
  __PutS__ := @S;
  PackString( Buffer, Put, H.BytesPerLine );
  Inc( H.YMax );
end;

destructor NewPCXPicture.Done;
var
  I:	Byte;
begin
  I := $0C;
  S.Write( I, 1 );
  S.Write( Pal, SizeOf(Pal) );
  S.Seek(0);
  S.Write( H, SizeOf(H) );
  S.Done;
end;

end.