{$I DEBUG}
unit Poly2D;

interface

uses Objects;

const
  CosSteps	= 256;
  CosScale	= 6;

type
  TLong = record
    case Word of
      1:(Lo,Hi:	Integer);
      2:(Val:	LongInt);
  end;

  PPoint2D = ^TPoint2D;
  TPoint2D = object(TObject)
    X,Y:  	TLong;
    Color:	Byte;
    constructor Init( AX, AY: LongInt);
    procedure Draw( Buffer: Pointer ); virtual;
    procedure Move( DX, DY: Integer ); virtual;
    procedure Rotate( CX, CY: Integer; R: Integer ); virtual;
    procedure Scale( CX,CY: Integer; RX, RY: Extended);
    procedure MorphTo( Point1, Point2: PPoint2D; NStep, NSteps: LongInt );
  end;

  PEdge = ^TEdge;
  TEdge = object(TObject)
    A,B: PPoint2D;
    CurX,
    Step: TLong;
    constructor Init( _A, _B: PPoint2D );
    procedure PrepareToDraw;
  end;

  TFLineDraw = procedure( Y, X1, X2: Integer; Color:Byte; Buffer:Pointer );
  PFLineDraw = ^TFLineDraw;

  PPoly2D = ^TPoly2D;
  TPoly2D = object(TPoint2D)
    Points,
    Edges: TCollection;
    PLine: PFLineDraw;
    constructor Init( APLine: PFLineDraw; AColor: Byte );
    procedure LoadDXF( Name: String );
    procedure InsertEdge( Edge: PEdge );
    procedure Draw( Buffer: Pointer ); virtual;
    procedure Move( DX, DY: Integer); virtual;
    procedure Rotate( CX, CY: Integer; R: Integer ); virtual;
    procedure Scale( CX,CY: Integer; RX,RY: Extended); virtual;
    procedure MorphTo( Poly1, Poly2: PPoly2D; NStep, NSteps: Integer );
    procedure Clear;
    destructor Done; virtual;
  end;

var
  CosTable,
  SinTable:	array[0..CosSteps-1] of LongInt;

implementation

uses VGAGraph, RStream;

constructor TPoint2D.Init( AX, AY: LongInt );
begin
  inherited Init;
  X.Val := AX;
  Y.Val := AY;
end;

procedure TPoint2D.Draw;
begin
{  PutPixelX( X.Hi, Y.Hi, Color );}
end;

procedure TPoint2D.Move( DX, DY: Integer );
begin
  Inc( X.Hi, DX ); Inc( Y.Hi, DY );
end;

procedure TPoint2D.Rotate( CX, CY: Integer; R: Integer );
var
  TTX, TTY,
  TX, TY:	LongInt;
begin
  if R < 0 then R := CosSteps+R;
  TX := X.Val-LongInt(CX) shl 16;
  TY := Y.Val-LongInt(CY) shl 16;
  if R > 0 then
  begin
{
    TTX := Round( (1.0*CosTable[R]*TX+1.0*SinTable[R]*TY)/(1 shl CosScale) );
    TTY := Round( (1.0*CosTable[R]*TY-1.0*SinTable[R]*TX)/(1 shl CosScale) );
}

    asm
      mov si,R
      shl si,2
      db $66; mov ax,word ptr CosTable[si]
      db $66; push ax
      db $66; cwd
      db $66; mul word ptr TX
      db $66; mov bx,ax
      db $66; mov ax,word ptr SinTable[si]
      db $66; push ax
      db $66; cwd
      db $66; mul word ptr TY
      db $66; sub bx,ax
      db $66; sar bx,CosScale
      db $66; pop ax
      db $66; cwd
      db $66; mul word ptr TX
      db $66; mov word ptr TX,bx
      db $66; mov bx,ax
      db $66; pop ax
      db $66; cwd
      db $66; mul word ptr TY
      db $66; add bx,ax
      db $66; sar bx,CosScale
      db $66; mov word ptr TY,bx
    end;
  end;
{
  X.Val := TTX+LongInt(CX) shl 16;
  Y.Val := TTY+LongInt(CY) shl 16;
}
  X.Val := TX+LongInt(CX) shl 16;
  Y.Val := TY+LongInt(CY) shl 16;
end;

constructor TPoly2D.Init( APLine: PFLineDraw; AColor: Byte );
begin
  Edges.Init(1,1);
  Points.Init(1,1);
  Color := AColor;
  PLine := APLine;
end;

procedure TPoly2D.LoadDXF( Name: String );

var
  F:		TRStream;
  S:		String;
  NewLine:	Boolean;
  P1, P2:	PPoint2D;

procedure GetNextEdge;
var
  Tmp:	Integer;
  X, Y:	Extended;

procedure GetNextPoint;
const
  PS : PPoint2D = nil;
begin
  P1 := P2;
  while (S <> 'VERTEX') and (F.Status = stOk) do
  begin
    if S = 'POLYLINE' then NewLine := True;
    S := GetLine(F);
  end;
  if F.Status <> stOk then Exit;
  GetLine(F);
  GetLine(F);
  GetLine(F);
  S := GetLine(F);
  Val( S, X, Tmp );
  GetLine(F);
  S := GetLine(F);
  Val( S, Y, Tmp );
  GetLine(F);
  S := GetLine(F);
  if S = 'SEQEND' then
    P2 := PS
  else
  begin
    P2 := New( PPoint2D, Init( Round((X+HRes div 2)*65536.0),
			       Round((VRes div 2 - Y)*65536.0) ) );
    if NewLine then PS := P2;
    if F.Status = stOk then Points.Insert(P2);
  end;
end;

begin
  GetNextPoint;
  if F.Status <> stOk then Exit;
  if NewLine = True then
  begin
    NewLine := False;
    GetNextPoint;
  end;
end;

begin
  F.Init( Name );
  GetNextEdge;
  repeat
    if F.Status <> stOk then Break;
    InsertEdge( New( PEdge, Init( P1, P2 ) ) );
    GetNextEdge;
  until F.Status <> stOk;
  F.Done;
end;

procedure TPoly2D.InsertEdge( Edge: PEdge );
begin
  Edges.Insert( Edge );
end;

type
  TEdgeACollection = object(TSortedCollection)
    function Compare(Key1, Key2: Pointer): Integer; virtual;
  end;

  TEdgeBCollection = object(TSortedCollection)
    function Compare(Key1, Key2: Pointer): Integer; virtual;
  end;

  TEdgeXCollection = object(TCollection)
    function Compare(Key1, Key2: Pointer): Integer;
    procedure Sort;
  end;

constructor TEdge.Init( _A, _B: PPoint2D );
begin
  inherited Init;
  A := _A; B := _B;
end;

procedure TEdge.PrepareToDraw;
var
  Tmp:	PPoint2D;
begin
  if A^.Y.Val > B^.Y.Val then
  begin
    Tmp := A; A := B; B := Tmp;
  end;
  CurX := A^.X;
  if B^.Y.Hi - A^.Y.Hi <> 0 then
    Step.Val := (B^.X.Val - A^.X.Val) div (B^.Y.Hi - A^.Y.Hi)
  else
    Step.Val := 0;
  if A^.Y.Val < 0 then
  begin
    Inc( CurX.Val, -A^.Y.Hi*Step.Val );
  end;
end;

procedure TPoint2D.MorphTo( Point1, Point2: PPoint2D; NStep, NSteps: LongInt );
var
  Tmp1,
  Tmp2:	LongInt;
begin
{
  X.Val := (Point2^.X.Val-Point1^.X.Val) div NSteps*NStep+Point1^.X.Val;
  Y.Val := (Point2^.Y.Val-Point1^.Y.Val) div NSteps*NStep+Point1^.Y.Val;
}

  Tmp2 := Point1^.X.Val;
  Tmp1 := Point2^.X.Val;
  asm
    push ds
    lds di,[bp+6]
    db $66; mov ax,word ptr Tmp1
    db $66; sub ax,word ptr Tmp2
    db $66; mul word ptr NStep
    db $66; cwd
    db $66; idiv word ptr NSteps
    db $66; add ax,word ptr Tmp2
    db $66; mov word ptr di.X.Val,ax
  end;
  Tmp2 := Point1^.Y.Val;
  Tmp1 := Point2^.Y.Val;
  asm
    lds di,[bp+6]
    db $66; mov ax,word ptr Tmp1
    db $66; sub ax,word ptr Tmp2
    db $66; mul word ptr NStep
    db $66; cwd
    db $66; idiv word ptr NSteps
    db $66; add ax,word ptr Tmp2
    db $66; mov word ptr di.Y.Val,ax
    pop ds
  end;
end;

function TEdgeACollection.Compare(Key1, Key2: Pointer): Integer;
begin
  Compare := 0;
  if PEdge(Key1)^.A^.Y.Hi < PEdge(Key2)^.A^.Y.Hi  then Compare := -1;
  if PEdge(Key1)^.A^.Y.Hi > PEdge(Key2)^.A^.Y.Hi  then Compare := 1;
end;

function TEdgeBCollection.Compare(Key1, Key2: Pointer): Integer;
begin
  Compare := 0;
  if PEdge(Key1)^.B^.Y.Hi < PEdge(Key2)^.B^.Y.Hi  then Compare := -1;
  if PEdge(Key1)^.B^.Y.Hi > PEdge(Key2)^.B^.Y.Hi  then Compare := 1;
end;

function TEdgeXCollection.Compare(Key1, Key2: Pointer): Integer;
begin
  Compare := 0;
  if (Key1 <> nil) and (Key2 <> nil) then
  begin
    if PEdge(Key1)^.CurX.Hi < PEdge(Key2)^.CurX.Hi  then Compare := -1;
    if PEdge(Key1)^.CurX.Hi > PEdge(Key2)^.CurX.Hi  then Compare := 1;
  end;
end;

procedure TEdgeXCollection.Sort;

procedure QSort(l, r: Integer);
var
  i, j: Integer;
  x, y: Pointer;
begin
  i := l; j := r; x := Items^[(l+r) DIV 2];
  repeat
    while Compare( Items^[i], x ) = -1 do i := i + 1;
    while Compare( x, Items^[j] ) = -1 do j := j - 1;
    if i <= j then
    begin
      y := Items^[i]; Items^[i] := Items^[j]; Items^[j] := y;
      i := i + 1; j := j - 1;
    end;
  until i > j;
  if l < j then QSort(l, j);
  if i < r then QSort(i, r);
end;

begin
  if Count > 1 then QSort( 0, Count-1 );
end;

procedure TPoly2D.Draw( Buffer: Pointer );

var
  EdgesA:	TEdgeACollection;
  EdgesB:	TEdgeBCollection;
  ScanLine:	TEdgeXCollection;
  I, J, K, L:	Integer;
  Tmp,
  Tmp1:		PEdge;
  VMin, VMax:	Integer;

begin
  EdgesA.Init( Edges.Count, 1 );
  EdgesB.Init( Edges.Count, 1 );
  EdgesA.Duplicates := True;
  EdgesB.Duplicates := True;
  VMin := 32767; VMax := -32767;
  for I := 1 to Edges.Count do
  begin
    Tmp := Edges.Items^[I-1];
    Tmp^.PrepareToDraw;
    if (Tmp^.B^.Y.Hi < 0) or (Tmp^.A^.Y.Hi >= VRes) then Continue;
    if Tmp^.A^.Y.Hi < VMin then VMin := Tmp^.A^.Y.Hi;
    if Tmp^.B^.Y.Hi > VMax then VMax := Tmp^.B^.Y.Hi;
    EdgesA.Insert(Tmp);
    EdgesB.Insert(Tmp);
  end;

  if VMin < 0 then VMin := 0;
  if VMax >= VRes then VMax := VRes - 1;

  ScanLine.Init( 1, 1 );
  K := 0; L := 0;
  for I := VMin to VMax do
  begin
    if K < EdgesA.Count then
    begin
      Tmp := PEdge(EdgesA.Items^[K]);
      while (Tmp^.A^.Y.Hi <= I) do
      begin
	ScanLine.Insert( Tmp );
	Inc(K);
	if K >= EdgesA.Count then Break;
	Tmp := PEdge(EdgesA.Items^[K]);
      end;
    end;
    if L < EdgesB.Count then
    begin
      Tmp := PEdge(EdgesB.At(L));
      while (Tmp^.B^.Y.Hi = I) do
      begin
	ScanLine.Delete( Tmp );
	Inc(L);
	if L >= EdgesB.Count then Break;
	Tmp := PEdge(EdgesB.At(L));
      end;
    end;

   if ScanLine.Count mod 2 <> 0 then
{      Continue;}
      Halt;
{    DrawHLineX( I, 0, Tmp1^.CurX.Hi, 0 );}
    J := 1;
    ScanLine.Pack;
    ScanLine.Sort;
    Tmp1 := PEdge(ScanLine.Items^[0]);
    repeat
      Tmp := Tmp1;
      Tmp1 := PEdge(ScanLine.Items^[J]);
      PLine^( I, Tmp^.CurX.Hi, Tmp1^.CurX.Hi, Color, Buffer );
{
      DrawHLineX( I, Tmp^.CurX.Hi, Tmp^.CurX.Hi+1, Color );
      DrawHLineX( I, Tmp1^.CurX.Hi, Tmp1^.CurX.Hi+1, Color );
}
      Inc( Tmp^.CurX.Val, Tmp^.Step.Val );
      Inc(J);
      if J >= ScanLine.Count then break;
      Tmp := Tmp1;
      Tmp1 := PEdge(ScanLine.Items^[J]);
{      DrawHLineX( I, Tmp^.CurX.Hi, Tmp1^.CurX.Hi, 0 );}
      Inc( Tmp^.CurX.Val, Tmp^.Step.Val );
      Inc(J);
    until False;
{    DrawHLineX( I, Tmp1^.CurX.Hi, HRes-1, 0 );}
    Inc( Tmp1^.CurX.Val, Tmp1^.Step.Val );
  end;
  EdgesA.DeleteAll;
  EdgesA.Done;
  EdgesB.DeleteAll;
  EdgesB.Done;
  ScanLine.DeleteAll;
  ScanLine.Done;
end;

procedure TPoly2D.Move( DX, DY: Integer);
var
  I :	Integer;
begin
  for I := 1 to Edges.Count do
  begin
    PEdge(Edges.At(I-1))^.A^.Move( DX,DY );
    PEdge(Edges.At(I-1))^.B^.Move( DX,DY );
  end;
end;

procedure TPoly2D.Rotate( CX, CY: Integer; R: Integer );
var
  I :	Integer;
begin
  for I := 1 to Edges.Count do
  begin
    PEdge(Edges.At(I-1))^.A^.Rotate( CX,CY,R );
    PEdge(Edges.At(I-1))^.B^.Rotate( CX,CY,R );
  end;
end;

procedure TPoly2D.MorphTo( Poly1, Poly2: PPoly2D; NStep, NSteps: Integer );
var
  I :	Integer;
begin
{
  for I := 0 to Edges.Count-1 do
  begin
    PEdge(Edges.Items^[I])^.A^.MorphTo(
      PEdge(Poly1^.Edges.Items^[I])^.A,
      PEdge(Poly2^.Edges.Items^[I])^.A,NStep,NSteps );
    PEdge(Edges.Items^[I])^.B^.MorphTo(
      PEdge(Poly1^.Edges.Items^[I])^.B,
      PEdge(Poly2^.Edges.Items^[I])^.B,NStep,NSteps );
  end;
}
  for I := 0 to Points.Count-1 do
  begin
    PPoint2D(Points.Items^[I])^.MorphTo(
      PPoint2D(Poly1^.Points.Items^[I]),
      PPoint2D(Poly2^.Points.Items^[I]),NStep,NSteps );
  end;
end;

procedure TPoint2D.Scale( CX,CY: Integer; RX, RY: Extended);
var
  I :	Integer;
  NCX,
  NCY:	TLong;
begin
  NCX.Hi := CX; NCX.Lo := 0;
  NCY.Hi := CY; NCY.Lo := 0;
  X.Val := Round((X.Val-NCX.Val)*RX)+NCX.Val;
  Y.Val := Round((Y.Val-NCY.Val)*RY)+NCY.Val;
{
  asm
    push ds
    lds di,[bp+6]
    db $66; mov ax,word ptr di.X.Val
    db $66; sub ax,word ptr NCX.Val
    db $66; cwd
    db $66; xor bx,bx
    mov bl,RX
    add bl,128
    db $66; mul bx
    db $66; sar ax,7
    db $66; add ax,word ptr NCX.Val
    db $66; mov word ptr di.X.Val,ax
    db $66; mov ax,word ptr di.Y.Val
    db $66; sub ax,word ptr NCY.Val
    db $66; cwd
    db $66; xor bx,bx
    mov bl,RY
    add bl,128
    db $66; mul bx
    db $66; sar ax,7
    db $66; add ax,word ptr NCY.Val
    db $66; mov word ptr di.Y.Val,ax
    pop ds
  end;
}
end;

procedure TPoly2D.Scale( CX,CY: Integer; RX, RY: Extended);
var
  I :	Integer;
begin
  for I := 1 to Points.Count do
  begin
    PPoint2D(Points.At(I-1))^.Scale( CX,CY, RX,RY );
  end;
end;

procedure TPoly2D.Clear;
begin
  Points.FreeAll;
  Edges.DeleteAll;
end;

destructor TPoly2D.Done;
begin
  Points.Done;
  Edges.DeleteAll;
  Edges.Done;
end;

var
  I:	Integer;

begin
  for I := 0 to CosSteps-1 do
  begin
    CosTable[I] := Round( Cos( 2*Pi*I/CosSteps )*(1 shl CosScale) );
    SinTable[I] := Round( Sin( 2*Pi*I/CosSteps )*(1 shl CosScale) );
  end;
end.