{$I DEBUG}
unit Polygon;

interface

{$Define VLines}
{$Define Gouraud}

uses Objects, VgaGraph;

const
  CosSteps	= 256;


type
  PPointer = ^Pointer;

  TLong = record
    Lo,
    Hi:	Word;
  end;

  P3DObject = ^T3DObject;
  T3DObject = object(TObject)
    SortValue:	LongInt;
    procedure InitTransform; virtual;
    procedure DoneTransform; virtual;
{$IfDef Gouraud}
    procedure ComputeBrightness; virtual;
{$EndIf}
    procedure Rotate( CX, CY, CZ: LongInt; RX, RY, RZ: Integer ); virtual;
    procedure Draw( Buffer: Pointer ); virtual;
    procedure Move( DX, DY, DZ: LongInt ); virtual;
    procedure ComputeSortValue; virtual;
    procedure GetScreenExtent( var Bounds: TRect ); virtual;
    function Visible: Boolean;  virtual;
end;

  PVertice = ^TVertice;
  TVertice = object(T3DObject)
{$IfDef Gouraud}
    SColor:	Integer;
    Divisor:	Integer;
{$EndIf}
    Transformed:Boolean;
    X,Y,Z:	LongInt;
    constructor Init( AX, AY, AZ: LongInt );
    constructor Load(var S: TStream);
    procedure InitTransform; virtual;
    procedure DoneTransform; virtual;
    procedure Move( DX, DY, DZ: LongInt ); virtual;
    procedure Rotate( CX, CY, CZ: LongInt; RX, RY, RZ: Integer ); virtual;
    procedure GetScreenExtent( var Bounds: TRect );  virtual;
    procedure Store(var S: TStream);
  end;

  P3DCollection = ^T3DCollection;
  T3DCollection = object(TCollection)
    function Compare(Key1, Key2: Pointer): Integer;
    procedure Sort;
  end;

  PCompoundObject = ^TCompoundObject;
  TCompoundObject = object(T3DObject)
    Points:	TCollection;
    Parts:	T3DCollection;
    constructor Init;
    constructor Load(var S: TStream);
    procedure Import3DS( FileName: String; Scale: Integer; Color: Byte );
    procedure InitTransform; virtual;
    procedure DoneTransform; virtual;
{$IfDef Gouraud}
    procedure ComputeBrightness; virtual;
{$EndIf}
    procedure InsertObject( AObject : P3DObject );
    procedure Draw( Buffer: Pointer ); virtual;
    procedure Move( DX, DY, DZ: LongInt ); virtual;

    procedure Rotate( CX, CY, CZ: LongInt;
		      RX, RY, RZ: Integer ); virtual;
    procedure ComputeSortValue; virtual;
    procedure GetScreenExtent( var Bounds: TRect ); virtual;
    procedure Store(var S: TStream);
    destructor Done; virtual;
  end;

  PFace3 = ^TFace3;
  TFace3 = object(T3DObject)
    Vertice:	array[1..3] of PVertice;
{$IfDef Gouraud}
    Normal:	TVertice;
{$EndIf}
    Color:	Byte;
    constructor Init( V1, V2, V3: PVertice; AColor: Byte );
    procedure InitTransform; virtual;
    procedure DoneTransform; virtual;
{$IfDef Gouraud}
    procedure ComputeBrightness; virtual;
{$EndIf}
    procedure Draw( Buffer: Pointer ); virtual;
    procedure Move( DX, DY, DZ: LongInt ); virtual;
    procedure Rotate( CX, CY, CZ: LongInt; RX, RY, RZ: Integer ); virtual;
    procedure ComputeSortValue; virtual;
    procedure GetScreenExtent( var Bounds: TRect ); virtual;
    function Visible: Boolean; virtual;
    destructor Done; virtual;
private
    procedure Sort;
  end;

  PFace4 = ^TFace4;
  TFace4 = object(T3DObject)
    Vertice:	array[1..4] of PVertice;
    Color:	Byte;
    Texture:	Pointer;
{$IfDef Gouraud}
    Normal:	TVertice;
{$EndIf}
    constructor InitColor( V1, V2, V3, V4: PVertice; AColor: Byte );
    constructor InitTexture( V1, V2, V3, V4: PVertice; ATexture: Pointer );
    procedure InitTransform; virtual;
    procedure DoneTransform; virtual;
{$IfDef Gouraud}
    procedure ComputeBrightness; virtual;
{$EndIf}
    procedure Draw( Buffer: Pointer ); virtual;
    procedure Move( DX, DY, DZ: LongInt ); virtual;
    procedure Rotate( CX, CY, CZ: LongInt; RX, RY, RZ: Integer ); virtual;
    procedure ComputeSortValue; virtual;
    procedure GetScreenExtent( var Bounds: TRect ); virtual;
    function Visible: Boolean; virtual;
    destructor Done; virtual;
  end;
{
  PPolygon = ^TPolygon;
  TPolygon = object(TObject)
    Vertice:	TCollection;
    Color:	Byte;
    constructor Init( AColor: Byte );
    procedure InsertVertice( AVertice: TVertice );
    procedure Draw; virtual;
    procedure Move( DX, DY, DZ: LongInt );
    procedure Rotate( var Center: TVertice; RX, RY, RZ: Integer );
    function SortValue: Integer; virtual;
    destructor Done; virtual;
  end;
}


implementation

uses Gouraud, MapToRct, RStream;

var
  CosTable,
  SinTable:	array[0..CosSteps-1] of LongInt;
  ArcCosTable:	array[0..255] of Byte;

procedure DrawVLine( X, Y1, Y2 : Integer; Color: Byte ); assembler;
asm
  mov cx,X
  and cx,3
  mov ax,$102
  shl ah,cl
  mov dx,$3C4
  out dx,ax

  mov ax,Y1
  mov bx,Y2
  cmp ax,bx
  jb @no_swap
  xchg ax,bx
@no_swap:
  test ax,8000h
  je @no_top_clip
  xor ax,ax
@no_top_clip:
  cmp bx,VRes
  jb @no_bot_clip
  mov bx,VRes
@no_bot_clip:
  mov cx,bx
  sub cx,ax
  jz @no_draw
  mov dx,$A000
  mov es,dx
  mov di,CurOffset
  mov bx,HRes
  shr bx,2
  mul bx
  add di,ax
  mov ax,X
  shr ax,2
  add di,ax
  mov al,Color
@vloop:
  mov es:[di],al
  add di,bx
  loop @vloop
@no_draw:
end;

{ Modified from public domain code by Dirk Hoeschen( 2:2445/110.1) }
procedure DrawHLine( Y, X1, X2 : Integer; Color: Byte );
begin
  asm
    mov   ax, $A000    {A000 nach AX}
    mov   es, ax       {Adresssegment nach ES}
    cld                {Richtungsflag lschen}
    mov   ax, HRes    {Anzahl der bytes in x-richtung}
    shr   ax,2
    mul   Y            {Multipliziert mit y}
    mov   di, ax       {base of scan line}
    add   di, CurOffset
    mov   bx, X1        {X nach BX}
    mov   dx, X2       {X2 nach DX}
    cmp   dx, bx       {Breite in Bytes x2-x}
    jg    @bigger      {ist x2 > X}
    mov   bx, X2       {X2 nach BX}
    mov   dx, X1        {X  nach DX}
  @bigger:
    test  bx,8000h
    jz	  @no_l_clip
    xor	  bx,bx
@no_l_clip:
    cmp   dx,HRes
    jb	  @no_r_clip
    mov   dx,HRes-1
@no_r_clip:
    cmp   bx,dx
    je    @EXIT
    mov   cl, bl       {veschieben nach lowbyte CX}
    shr   bx, 2        {durch 4 teilen 1Punkt 4 Planes}
    mov   ch, dl       {verschieben nach hibyte cx}
    shr   dx, 2        {durch 4 teilen}
    sub   dx, bx       {Breite in Bytes x2-x}
    and   cx, $0303    {Jeweils die ersten 2 Bit HI und low Ausmaskieren}
    add   di, bx       {Zeilenoffset auf DI}
    mov   ax, $FF02    {Hibyte fllen lowbyte bit 1 setzen}
    shl   ah, cl       {hibyte um anfangsoffset shiften}
    and   ah, $0F      {linken Rand ausmaskieren}
    mov   cl, ch       {rechter offset nach cl}
    mov   bh, $F1      {11110001 nach Bh}
    rol   bh, cl       {rollen im Hibyte um ch}
    and   bh, $0F      {rechten rand ausmaskieren}
    mov   cx, dx       {breite in bytes nach cx}
    or    cx, cx       {was soll das denn?}
    jnz   @LEFT        {achso testen auf null}
    and   ah, bh       {Kombiniere linke & rechte bitmaske}
  @LEFT:
    mov   dx, $3C4     {Auswahlport fuer die planes}
    out   dx, ax       {planes auswaehlen}
    inc   dx           {nach schreibport erhoehen 3d5}
    mov   al, color    {farbe nach al}
    stosb              {und linken rand schreiben}
    jcxz  @EXIT        {wenn cx null ist dann abbrechen}
    dec   cx           {ansonsten decrementieren}
    jcxz  @RIGHT       {wenn null, dann rechte seite schreiben}
    mov   al, $0F      {00001111 alle planes auswaehlen}
    out   dx, al       {skipped if cx=0,1}
    mov   al, color    {Farbe nach al}
    repz  stosb        {mittlere bytes schreiben jeweils 4 Punkte}
   @RIGHT:
    mov   al, bh       {rechte maske nach al}
    out   dx, al       {skipped if cx=0}
    mov   al, color    {Farbe nach al}
    stosb              {rechte maske schreiben}
   @EXIT:
  end;
end;

{ * T3DObject * }

procedure T3DObject.Draw; begin Abstract end;
procedure T3DObject.InitTransform; begin Abstract end;
procedure T3DObject.DoneTransform; begin Abstract end;
{$IfDef Gouraud}
procedure T3DObject.ComputeBrightness; begin Abstract end;
{$EndIf}
procedure T3DObject.Move( DX, DY, DZ: LongInt ); begin Abstract end;
procedure T3DObject.Rotate( CX, CY, CZ:LongInt; RX, RY, RZ: Integer );
begin Abstract end;
procedure T3DObject.ComputeSortValue; begin Abstract end;
procedure T3DObject.GetScreenExtent; begin Abstract end;
function T3DObject.Visible: Boolean; begin Visible := True; end;

{ * T3DCollection * }

function T3DCollection.Compare(Key1, Key2: Pointer): Integer;
var
  V1, V2:	LongInt;
begin

  V1 := P3DObject(Key1)^.SortValue;
  V2 := P3DObject(Key2)^.SortValue;

  Compare := 0;

  if V1 > V2 then
    Compare := 1;
  if V1 < V2 then
    Compare := -1;

{
  if LongInt(Key1^) > LongInt(Key2^) then T3DCollection_Compare := -1;
  if LongInt(Key1^) < LongInt(Key2^) then T3DCollection_Compare := 1;
}
end;

procedure T3DCollection.Sort;

procedure QSort(l, r: Integer);
var
  i, j:	Integer;
  x, y:	Pointer;
begin
  i := l; j := r; x := Items^[(l+r) DIV 2];
  repeat
    while Compare( Items^[i], x ) = -1 do i := i + 1;
    while Compare( x, Items^[j] ) = -1 do j := j - 1;
    if i <= j then
    begin
      y := Items^[i]; Items^[i] := Items^[j]; Items^[j] := y;
      i := i + 1; j := j - 1;
    end;
  until i > j;
  if l < j then QSort(l, j);
  if i < r then QSort(i, r);
end;

begin
  QSort( 0, Count-1 );
end;

{ * TCompoundObject * }

constructor TCompoundObject.Init;
begin
  inherited Init;
  Points.init(1,1);
  Parts.Init(1,1);
{  Parts.Duplicates := True;}
end;

constructor TCompoundObject.Load(var S: TStream);
var
  A, B, C, CC,
  N,
  I:    Integer;
begin
  inherited Init;
  Points.Init(1,1);
  S.Read( N, SizeOf(N) );
  for I := 1 to N do
    Points.Insert( New( PVertice, Load(S) ) );
  S.Read( N, SizeOf(N) );
  Parts.Init(1,1);
  for I := 1 to N do
  begin
    S.Read( A, SizeOf(A) );
    S.Read( B, SizeOf(B) );
    S.Read( C, SizeOf(C) );
    S.Read( CC, SizeOf(CC) );
    Parts.Insert( New( PFace3, Init( Points.At(A),
				     Points.At(B),
                                     Points.At(C), CC ) ) );
  end;
end;

procedure TCompoundObject.Store(var S: TStream);
var
 I,
 J,
 Tmp:   Integer;
begin
  S.Write( Points.Count, SizeOf(Points.Count) );
  for I := 1 to Points.Count do
    PVertice( Points.At(I-1) )^.Store(S);
  S.Write( Parts.Count, SizeOf(Parts.Count) );
  for I := 1 to Parts.Count do
  begin
    for J := 1 to 3 do
    begin
      Tmp := Points.IndexOf( PFace3(Parts.At(I-1))^.Vertice[J] );
      S.Write( Tmp, SizeOf(Tmp) );
    end;
    Tmp := PFace3(Parts.At(I-1))^.Color;
    S.Write( Tmp, SizeOf(Tmp) );
  end;
end;

procedure TCompoundObject.Import3DS( FileName: String; Scale: integer;
				     Color: Byte );
const
  Spaces = [' ',#9,#10,#13,':',','];

var
  F:		TRStream;

function GetToken: String;
var
  S:	String;
  C:	Char;
begin
  S := '';
  repeat F.Read( C, 1 ) until not (C in Spaces) or (F.Status<>stOk);
  repeat
    S := S + C;
    F.Read( C, 1 )
  until (C in Spaces) or (F.Status<>stOk);
  GetToken := S;
end;

function GetLine: String;
var
  S:	String;
  C:	Char;
begin
  S := '';
  C := #0;
  repeat F.Read( C, 1 ) until not (C in Spaces) or (F.Status<>stOk);
  while (C <> #13) and (F.Status=stOk) do
  begin
    S := S + C;
    F.Read( C, 1 )
  end;
  GetLine := S;
end;

function GetNumber: Real;
var
  S:	String;
  Num:	Real;
  Code:	Integer;
begin
  S := GetToken;
  Val( S, Num, Code );
  GetNumber := Num;
end;

procedure WaitToken( Token: String );
var
  S:	String;
begin
  repeat S := GetToken until (S = Token) or (F.Status <> stOk);
end;

function FMax( A,B: Real): Real;
begin
  if A > B then FMax := A else FMax := B;
end;

function FMin( A,B: Real): Real;
begin
  if A < B then FMin := A else FMin := B;
end;

const
  MaxVertices = 2000;

var
  Vertices:	array[0..MaxVertices] of PVertice;
  X,Y,Z,
  MinX, MinY, MinZ,
  MaxX, MaxY, MaxZ,
  Max, Min:	Real;
  I,
  A,B,C,
  NFaces,
  NVertices:	Integer;
  P:		PCompoundObject;
  Name:		String;
  Start:	LongInt;

begin
  F.Init( FileName );
  if F.Status <> stOk then Exit;
  Start := F.GetPos;
  MinX := 999999;
  MinY := 999999;
  MinZ := 999999;
  MaxX := -999999;
  MaxY := -999999;
  MaxZ := -999999;
{  repeat}
    WaitToken( 'Vertices' );
{    if F.Status <> stOk then Break;}
    NVertices := Trunc(GetNumber);
    for I := 1 to NVertices do
    begin
      WaitToken( 'Vertex' );
      if F.Status <> stOk then Break;
      WaitToken( 'X' );
      if F.Status <> stOk then Break;
      X := GetNumber;
      MinX := FMin( X, MinX );
      MaxX := FMax( X, MaxX );
      WaitToken( 'Y' );
      Y := GetNumber;
      MinY := FMin( Y, MinY );
      MaxY := FMax( Y, MaxY );
      WaitToken( 'Z' );
      Z := GetNumber;
      MinZ := FMin( Z, MinZ );
      MaxZ := FMax( Z, MaxZ );
    end;
{  until False;}
  Max := FMax( MaxX-MinX, MaxY-MinY );
  Max := FMax( Max, MaxZ-MinZ );
  F.Reset;
  F.Seek(Start);
{  repeat}
    WaitToken( 'Named' );
    WaitToken( 'object' );
    Name := GetLine;
{    if GetToken <> 'Tri-mesh' then Continue;}
    if F.Status <> stOk then Exit;
    New( P, Init );
    WaitToken( 'Vertices' );
    NVertices := Trunc(GetNumber);
    if NVertices > MaxVertices then
    begin
      F.Done;
      Exit;
    end;
    WaitToken( 'Faces' );
    NFaces := Trunc(GetNumber);
    for I := 0 to NVertices-1 do
    begin
{      Write( 'Reading vertex #', I, '        '#13 );}
      WaitToken( 'X' );
      X := (GetNumber-(MaxX+MinX)/2)*Scale / (Max);
      WaitToken( 'Y' );
      Y := (GetNumber-(MaxY+MinY)/2)*Scale / (Max);
      WaitToken( 'Z' );
      Z := (GetNumber-(MaxZ+MinZ)/2)*Scale / (Max);
      New( Vertices[I], Init( Round(X), Round(-Z), Round(Y) ) );
      Points.Insert( Vertices[I] );
    end;
    for I := 0 to NFaces-1 do
    begin
{      Write( 'Reading face #', I, '        '#13 );}
      WaitToken( 'A' ); A := Trunc(GetNumber);
      WaitToken( 'B' ); B := Trunc(GetNumber);
      WaitToken( 'C' ); C := Trunc(GetNumber);

      InsertObject( New( PFace3,
	Init( Vertices[A], Vertices[B], Vertices[C], Color ) ) );

{
      if Name = 'face  ' then
	InsertObject( New( PFace3,
	  Init( Vertices[A], Vertices[B], Vertices[C], 0 ) ) )
      else
	InsertObject( New( PFace3,
	  Init( Vertices[A], Vertices[B], Vertices[C], 64 ) ) )
}
    end;
{  until F.Status <> stOk;}
end;

{$IfDef Gouraud}
procedure TCompoundObject.ComputeBrightness;
var
  I:	Integer;
begin
  for I := 1 to Parts.Count do
    T3DObject(Parts.Items^[I-1]^).ComputeBrightness;
end;
{$EndIf}

procedure TCompoundObject.InitTransform;
var
  I:	Integer;
begin
  for I := 1 to Parts.Count do
    T3DObject(Parts.Items^[I-1]^).InitTransform;
end;

procedure TCompoundObject.DoneTransform;
var
  I:	Integer;
begin
  for I := 1 to Parts.Count do
    T3DObject(Parts.Items^[I-1]^).DoneTransform;
{  ComputeSortValue;}
end;

procedure TCompoundObject.InsertObject( AObject : P3DObject );
begin
  if AObject <> nil then
    Parts.Insert( AObject );
end;

procedure TCompoundObject.Draw( Buffer: Pointer );
var
  I,
  J,
  Tmp:		Integer;
  P:		P3DObject;
  Bounds:	TRect;
  ToDraw:       TCompoundObject;
begin
  ToDraw.Init;
  ToDraw.Parts.SetLimit( Parts.Count+1 );
  InitTransform;
  for I := 1 to Parts.Count do
  begin
    P := Parts.Items^[I-1];
    if P^.Visible then ToDraw.InsertObject( P );
  end;
  ToDraw.ComputeSortValue;
  ToDraw.Parts.Sort;
  for I := 1 to {ToDraw.}Parts.Count do
  begin
    P := P3DObject({ToDraw.}Parts.Items^[I-1]);
    P^.ComputeBrightness;
  end;
  for I := 1 to ToDraw.Parts.Count do
    P3DObject(ToDraw.Parts.Items^[I-1])^.Draw( Buffer );
  ToDraw.Parts.DeleteAll;
  ToDraw.Done;
end;

procedure TCompoundObject.GetScreenExtent( var Bounds: TRect );
var
  I,J:	Integer;
  P:	P3DObject;
  Tmp:	TRect;
begin
  for I := 1 to Parts.Count do
  begin
    P := P3DObject(Parts.Items^[I-1]);
    for J := 1 to 3 do
    begin
      P^.GetScreenExtent( Tmp );
      if Tmp.A.X < Bounds.A.X then Bounds.A.X := Tmp.A.X;
      if Tmp.A.Y < Bounds.A.Y then Bounds.A.Y := Tmp.A.Y;
      if Tmp.B.X > Bounds.B.X then Bounds.B.X := Tmp.B.X;
      if Tmp.B.Y > Bounds.B.Y then Bounds.B.Y := Tmp.B.Y;
    end;
    P^.ComputeBrightness;
  end;
end;

procedure TCompoundObject.Move( DX, DY, DZ: LongInt );
var
  I:	Integer;
begin
  for I := 1 to Parts.Count do
    P3DObject(Parts.Items^[I-1])^.Move( DX, DY, DZ );
end;

procedure TCompoundObject.Rotate;
var
  I:	Integer;
begin
{  InitTransform;}
  for I := 0 to Points.Count-1 do
    PVertice(Points.Items^[I])^.Rotate( CX, CY, CZ, RX, RY, RZ );
  for I := 0 to Parts.Count-1 do
    PFace3(Parts.Items^[I])^.Normal.Rotate( CX, CY, CZ, RX, RY, RZ );
  DoneTransform;
end;

procedure TCompoundObject.ComputeSortValue;
var
  I:	Integer;
  Tmp:	P3DObject;
begin
  SortValue := 0;
  for I := 1 to Parts.Count do
  begin
    Tmp := P3DObject(Parts.At(I-1));
    Tmp^.ComputeSortValue;
    Inc( SortValue, Tmp^.SortValue );
  end;
  if Parts.Count > 0 then SortValue := SortValue div Parts.Count;
end;

destructor TCompoundObject.Done;
begin
  Points.FreeAll;
  Points.Done;
  Parts.Done;
  inherited Done;
end;

{ * TVertice * }

constructor TVertice.Init( AX, AY, AZ: LongInt );
begin
  inherited Init;
  X := AX; Y := AY; Z := AZ; Transformed := False;
  SColor := 0;
  Divisor := 0;
end;

constructor TVertice.Load(var S: TStream);
begin
  inherited Init;
  S.Read( X, SizeOf(X) );
  S.Read( Y, SizeOf(Y) );
  S.Read( Z, SizeOf(Z) );
  Transformed := True;
end;

procedure TVertice.Store(var S: TStream);
begin
  S.Write( X, SizeOf(X) );
  S.Write( Y, SizeOf(Y) );
  S.Write( Z, SizeOf(Z) );
end;

procedure TVertice.InitTransform;
begin
  SColor := 0;
  Divisor := 0;
end;

procedure TVertice.DoneTransform;
begin
  Transformed := False;
end;

procedure TVertice.Move( DX, DY, DZ: LongInt );
begin
  if not Transformed then
  begin
    Inc( X, DX shl 8 ); Inc( Y, DY shl 8 ); Inc( Z, DZ shl 8 );
    Transformed := True;
  end;
end;

procedure TVertice.Rotate( CX, CY, CZ:LongInt; RX, RY, RZ: Integer );
var
  TX, TY, TZ,
  T1, T2, T3, T4:	LongInt;
begin
  if not Transformed then
  begin
    TX := X-CX;
    TY := Y-CY;
    TZ := Z-CZ;
    if RY <> 0 then
    begin
      if RY < 0 then RY := RY mod CosSteps + CosSteps;
      asm
	mov si,RY
	shl si,2
	db $66; mov ax,word ptr CosTable[si]
	db $66; push ax
	db $66; cwd
	db $66; mul word ptr TX
	db $66; mov bx,ax
	db $66; mov ax,word ptr SinTable[si]
	db $66; push ax
	db $66; cwd
	db $66; mul word ptr TZ
	db $66; sub bx,ax
	db $66; sar bx,10
	db $66; pop ax
	db $66; cwd
	db $66; mul word ptr TX
	db $66; mov word ptr TX,bx
	db $66; mov bx,ax
	db $66; pop ax
	db $66; cwd
	db $66; mul word ptr TZ
	db $66; add bx,ax
	db $66; sar bx,10
	db $66; mov word ptr TZ,bx
      end;
{
      T1 := TX * CosTable[ RY ];
      T2 := TZ * SinTable[ RY ];
      T3 := TX * SinTable[ RY ];
      T4 := TZ * CosTable[ RY ];
      TX := ( T1 - T2 ) div 1024;
      TZ := ( T3 + T4 ) div 1024;
}
    end;
    if RX > 0 then
    begin
      if RX < 0 then RX := RX mod CosSteps + CosSteps;
     asm
	mov si,RX
	shl si,2
	db $66; mov ax,word ptr CosTable[si]
	db $66; push ax
	db $66; cwd
	db $66; mul word ptr TY
	db $66; mov bx,ax
	db $66; mov ax,word ptr SinTable[si]
	db $66; push ax
	db $66; cwd
	db $66; mul word ptr TZ
	db $66; sub bx,ax
	db $66; sar bx,10
	db $66; pop ax
	db $66; cwd
	db $66; mul word ptr TY
	db $66; mov word ptr TY,bx
	db $66; mov bx,ax
	db $66; pop ax
	db $66; cwd
	db $66; mul word ptr TZ
	db $66; add bx,ax
	db $66; sar bx,10
	db $66; mov word ptr TZ,bx
      end;
{
      T1 := TY * CosTable[ RX ];
      T2 := TZ * SinTable[ RX ];
      T3 := TY * SinTable[ RX ];
      T4 := TZ * CosTable[ RX ];
      TY := ( T1 - T2 ) div 1024;
      TZ := ( T3 + T4 ) div 1024;
}
    end;
    if RZ > 0 then
    begin
    if RZ < 0 then RZ := RZ mod CosSteps + CosSteps;
     asm
	mov si,RZ
	shl si,2
	db $66; mov ax,word ptr CosTable[si]
	db $66; push ax
	db $66; cwd
	db $66; mul word ptr TX
	db $66; mov bx,ax
	db $66; mov ax,word ptr SinTable[si]
	db $66; push ax
	db $66; cwd
	db $66; mul word ptr TY
	db $66; sub bx,ax
	db $66; sar bx,10
	db $66; pop ax
	db $66; cwd
	db $66; mul word ptr TX
	db $66; mov word ptr TX,bx
	db $66; mov bx,ax
	db $66; pop ax
	db $66; cwd
	db $66; mul word ptr TY
	db $66; add bx,ax
	db $66; sar bx,10
	db $66; mov word ptr TY,bx
      end;
{
      T1 := TY * CosTable[ RZ ];
      T2 := TX * SinTable[ RZ ];
      T3 := TY * SinTable[ RZ ];
      T4 := TX * CosTable[ RZ ];
      TY := ( T1 - T2 ) div 1024;
      TX := ( T3 + T4 ) div 1024;
}
    end;
    X := TX+CX;
    Y := TY+CY;
    Z := TZ+CZ;
    Transformed := True;
  end;
end;

procedure TVertice.GetScreenExtent( var Bounds: TRect );
begin
  Bounds.A.X := X shr 8; Bounds.B.X := X shr 8;
  Bounds.A.Y := Y shr 8; Bounds.A.X := Y shr 8;
end;

{ * TFace3 * }

constructor TFace3.Init( V1, V2, V3: PVertice; AColor: Byte );
var
  I:	Integer;
  N:	Real;
begin
  Vertice[1] := V1; Vertice[2] := V2; Vertice[3] := V3;
  for I := 1 to 3 do
  begin
    if not Vertice[I]^.Transformed then
    begin
      Vertice[I]^.Transformed := True;
      Vertice[I]^.X := Vertice[I]^.X shl 8;
      Vertice[I]^.Y := Vertice[I]^.Y shl 8;
      Vertice[I]^.Z := Vertice[I]^.Z shl 8;
    end;
  end;
  Normal.Init(((Vertice[2]^.Y-Vertice[1]^.Y)*(Vertice[3]^.Z-Vertice[1]^.Z) -
	       (Vertice[2]^.Z-Vertice[1]^.Z)*(Vertice[3]^.Y-Vertice[1]^.Y)) div 256,
	      ((Vertice[2]^.Z-Vertice[1]^.Z)*(Vertice[3]^.X-Vertice[1]^.X) -
	       (Vertice[2]^.X-Vertice[1]^.X)*(Vertice[3]^.Z-Vertice[1]^.Z)) div 256,
	      ((Vertice[2]^.X-Vertice[1]^.X)*(Vertice[3]^.Y-Vertice[1]^.Y) -
	       (Vertice[2]^.Y-Vertice[1]^.Y)*(Vertice[3]^.X-Vertice[1]^.X)) div 256 );
  N := Sqrt(Sqr(Normal.X div 256) + Sqr(Normal.Y div 256) + Sqr(Normal.Z div 256));
  if N = 0 then Fail;
  Normal.X := Round( Normal.X / N * 256 );
  Normal.Y := Round( Normal.Y / N * 256 );
  Normal.Z := Round( Normal.Z / N * 256 );
  Color := AColor;
  Sort;
end;

{$IfDef Gouraud}
procedure TFace3.ComputeBrightness;
var
  I:		Integer;
  A,B,C:	LongInt;
  XS,YS,ZS:	Integer;
begin
{
  A := Normal.X div 256;
  B := Normal.Y div 256;
  C := Normal.Z div 256;
}
  asm
    les di,[bp+06]
{
    db $66; mov ax, word ptr es:di[Normal.X]
    db $66; sar ax, 8
    db $66; mov word ptr A,ax
    db $66; mov ax, word ptr es:di[Normal.Y]
    db $66; sar ax, 8
    db $66; mov word ptr B,ax
}
    db $66; mov ax, word ptr es:di[Normal.Z]
    db $66; sar ax, 8
    db $66; mov word ptr C,ax
  end;
  for I := 1 to 3 do with Vertice[I]^ do
  begin
{
    XS := X shr 8;
    YS := Y shr 8;
    ZS := Z shr 8 - 150;
}
    if C > 0 then Inc( SColor, C );
    Inc( Divisor );
(*
    asm
      db $66; mov ax,word ptr es:di[X]
      db $66; sar ax,8
      push ax
      db $66; mul word ptr A
      db $66; mov cx,ax
      db $66; mov ax,word ptr es:di[Y]
      db $66; sar ax,8
      mov bx,ax
      db $66; mul word ptr B
      db $66; add cx,ax
      db $66; mov ax,word ptr es:di[Z]
      db $66; sar ax,8
      db $66; sub ax,250; dw 0
      mov si,ax
      db $66; mul word ptr C
      db $66; add cx,ax
      db $66,$0F,$BF,$C3	{movsx eax,bx}
      db $66; imul ax
      db $66; mov bx,ax
      pop ax
      db $66,$0F,$BF,$C0	{movsx eax,ax}
      db $66; imul ax
      db $66; add bx,ax
      db $66,$0F,$BF,$C6	{movsx eax,si}
      db $66; imul ax
      db $66; add bx,ax
      db $66; shr bx,2
      mov ax,es
      les dx,SqrtTable
      mov bl,es:[bx]
      xor bh,bh
      mov es,ax
      shl bx,1
      or bx,bx
      je @no_div
      db $66; mov ax,cx
      db $66,$99		{cdq}
      db $66; idiv bx
      test ax,$8000
      jz @no_neg
      neg ax
@no_neg:
      mov bx,ax
@no_div:
      or bh,bh
      je @no_trunc
      mov bx,255
@no_trunc:
      mov al,[offset ArcCosTable+bx]
      xor ah,ah
      add es:di[SColor],ax
      inc word ptr es:di[Divisor]
    end;
*)
  end;
end;
{$EndIf}

procedure TFace3.InitTransform;
var
  I:		Integer;
begin
  for I := 1 to 3 do Vertice[I]^.InitTransform;
end;

procedure TFace3.DoneTransform;
var
  I:	Integer;
begin
  for I := 1 to 3 do
    Vertice[I]^.DoneTransform;
  Normal.DoneTransform;
end;

{$IfDef HLines}

procedure TFace3.Draw( Buffer: Pointer );
var
  I,
  X,Y,
  X1,Y1,
  X2,Y2,
  X3,Y3:	Integer;
  Step1,
  Step2,
  XX1,
  XX2:		LongInt;
begin
  X1 := Vertice[1]^.X shr 8 + HRes div 2;
  Y1 := Vertice[1]^.Y shr 8 + VRes div 2;
  X2 := Vertice[2]^.X shr 8 + HRes div 2;
  Y2 := Vertice[2]^.Y shr 8 + VRes div 2;
  X3 := Vertice[3]^.X shr 8 + HRes div 2;
  Y3 := Vertice[3]^.Y shr 8 + VRes div 2;
  if (Y1 >= VRes) or (Y3 < 0) or (Y1=Y3) then Exit;
  XX1 := LongInt(X1) shl 16; XX2 := XX1;
  Step1 := (LongInt(X3-X1) shl 16) div (Y3-Y1);
  if Y1 <> Y2 then
  begin
    Step2 := (LongInt(X2-X1) shl 16) div (Y2-Y1);
    if Y1 < 0 then
    begin
      XX1 := XX1 - Y1*Step1;
      XX2 := XX2 - Y1*Step2;
      Y1 := 0;
    end;
    if Y2 > VRes then Y2 := VRes;
    while (Y1 < Y2) do
    begin
      if (TLong(XX1).Hi < HRes ) and  (TLong(XX2).Hi > 0 ) then
	DrawHLine( Y1, TLong(XX1).Hi, TLong(XX2).Hi, Y1 );
      Inc( XX1, Step1 );
      Inc( XX2, Step2 );
      Inc(Y1);
    end;
  end;
  if Y2 <> Y3 then
  begin
    XX2 := LongInt(X2) shl 16;
    Step2 := (LongInt(X3-X2) shl 16) div (Y3-Y2);
    if Y3 > VRes then Y3 := VRes;
    while (Y1 < Y3) do
    begin
      if (TLong(XX1).Hi < HRes ) and  (TLong(XX2).Hi > 0 ) then
	DrawHLine( Y1, TLong(XX1).Hi, TLong(XX2).Hi, Y1 );
      Inc( XX1, Step1 );
      Inc( XX2, Step2 );
      Inc(Y1);
    end;
  end;
end;

{$Else}
{$IfDef Gouraud}
procedure TFace3.Draw( Buffer: Pointer );
var
  Color1,
  Color2,
  Color3:	Integer;

begin
  Color1 := Vertice[1]^.SColor div Vertice[1]^.Divisor;
  Color2 := Vertice[2]^.SColor div Vertice[2]^.Divisor;
  Color3 := Vertice[3]^.SColor div Vertice[3]^.Divisor;

  DrawGouraudTriangle( Integer(Vertice[1]^.X shr 8)*6 div 5+HRes div 2, Vertice[1]^.Y shr 8+VRes div 2,
		       Integer(Vertice[2]^.X shr 8)*6 div 5+HRes div 2, Vertice[2]^.Y shr 8+VRes div 2,
		       Integer(Vertice[3]^.X shr 8)*6 div 5+HRes div 2, Vertice[3]^.Y shr 8+VRes div 2,
		       Color+Color1 shr 2, Color+Color2 shr 2, Color+Color3 shr 2,
		       Buffer );

end;
{$Else}
procedure TFace3.Draw( Buffer: Pointer );
var
  I,
  X,Y,
  X1,Y1,
  X2,Y2,
  X3,Y3:	Integer;
  Step1,
  Step2,
  YY1,
  YY2:		LongInt;
begin
  X1 := Vertice[1]^.X shr 8 + HRes div 2;
  Y1 := Vertice[1]^.Y shr 8 + VRes div 2;
  X2 := Vertice[2]^.X shr 8 + HRes div 2;
  Y2 := Vertice[2]^.Y shr 8 + VRes div 2;
  X3 := Vertice[3]^.X shr 8 + HRes div 2;
  Y3 := Vertice[3]^.Y shr 8 + VRes div 2;
  if (X1 >= HRes) or (X3 < 0) or (X1=X3) then Exit;
  YY1 := LongInt(Y1) shl 16; YY2 := YY1;
  Step1 := (LongInt(Y3-Y1) shl 16) div (X3-X1);
  if X1 <> X2 then
  begin
    Step2 := (LongInt(Y2-Y1) shl 16) div (X2-X1);
    while (X1 < X2) do
    begin
      DrawVLine( X1, TLong(YY1).Hi, TLong(YY2).Hi, Color );
      Inc( YY1, Step1 );
      Inc( YY2, Step2 );
      Inc(X1);
    end;
  end;
  if X2 <> X3 then
  begin
    YY2 := LongInt(Y2) shl 16;
    Step2 := (LongInt(Y3-Y2) shl 16) div (X3-X2);
    while (X1 < X3) do
    begin
      DrawVLine( X1, TLong(YY1).Hi, TLong(YY2).Hi, Color );
      Inc( YY1, Step1 );
      Inc( YY2, Step2 );
      Inc(X1);
    end;
  end;
end;

{$EndIf}
{$EndIf}

procedure TFace3.Move( DX, DY, DZ: LongInt );
begin
  Vertice[1]^.Move( DX, DY, DZ );
  Vertice[2]^.Move( DX, DY, DZ );
  Vertice[3]^.Move( DX, DY, DZ );
end;

procedure TFace3.Rotate( CX, CY, CZ:LongInt; RX, RY, RZ: Integer );
begin
  Vertice[1]^.Rotate( CX, CY, CZ, RX, RY, RZ );
  Vertice[2]^.Rotate( CX, CY, CZ, RX, RY, RZ );
  Vertice[3]^.Rotate( CX, CY, CZ, RX, RY, RZ );
  Normal.Rotate( CX, CY, CZ, RX, RY, RZ );
end;

procedure TFace3.ComputeSortValue;
begin
  SortValue := (Vertice[1]^.Z + Vertice[2]^.Z +Vertice[3]^.Z);
end;

destructor TFace3.Done;
begin
{
  Vertice[1]^.Free;
  Vertice[2]^.Free;
  Vertice[3]^.Free;
}
end;

{$IfDef HLines}

procedure TFace3.Sort;
var
  Tmp:	PVertice;
begin
  if Vertice[1]^.Y > Vertice[2]^.Y then
  begin
    Tmp := Vertice[1]; Vertice[1] := Vertice[2]; Vertice[2] := Tmp;
  end;
  if Vertice[2]^.Y > Vertice[3]^.Y then
  begin
    Tmp := Vertice[2]; Vertice[2] := Vertice[3]; Vertice[3] := Tmp;
  end;
  if Vertice[1]^.Y > Vertice[2]^.Y then
  begin
    Tmp := Vertice[1]; Vertice[1] := Vertice[2]; Vertice[2] := Tmp;
  end;
end;

{$Else}

{$IfDef Gouraud}
procedure TFace3.Sort;
begin
end;

{$Else}

procedure TFace3.Sort;
var
  Tmp:	PVertice;
begin
  if Vertice[1]^.X > Vertice[2]^.X then
  begin
    Tmp := Vertice[1]; Vertice[1] := Vertice[2]; Vertice[2] := Tmp;
  end;
  if Vertice[2]^.X > Vertice[3]^.X then
  begin
    Tmp := Vertice[2]; Vertice[2] := Vertice[3]; Vertice[3] := Tmp;
  end;
  if Vertice[1]^.X > Vertice[2]^.X then
  begin
    Tmp := Vertice[1]; Vertice[1] := Vertice[2]; Vertice[2] := Tmp;
  end;
end;

{$EndIf Gouraud}

{$EndIf}

function TFace3.Visible: Boolean;
begin
  Visible := Normal.Z > 0;
end;

procedure TFace3.GetScreenExtent( var Bounds: TRect );
var
  I:	Integer;
begin
  Bounds.A.X := 256; Bounds.A.Y := 256;
  Bounds.B.X := 0; Bounds.B.Y := 0;
  for I := 1 to 3 do
  begin
    if Vertice[I]^.X < Bounds.A.X then Bounds.A.X := Vertice[I]^.X;
    if Vertice[I]^.Y < Bounds.A.Y then Bounds.A.Y := Vertice[I]^.Y;
    if Vertice[I]^.X > Bounds.B.X then Bounds.B.X := Vertice[I]^.X;
    if Vertice[I]^.Y > Bounds.B.Y then Bounds.B.Y := Vertice[I]^.Y;
  end;
end;

{ * TFace4 * }

constructor TFace4.InitColor( V1, V2, V3, V4: PVertice; AColor: Byte );
begin
  Vertice[1] := V1; Vertice[2] := V2; Vertice[3] := V3; Vertice[4] := V4;
  Color := AColor; Texture := nil;
end;

constructor TFace4.InitTexture( V1, V2, V3, V4: PVertice; ATexture: Pointer );
begin
  Vertice[1] := V1; Vertice[2] := V2; Vertice[3] := V3; Vertice[4] := V4;
  Texture := ATexture;
end;

procedure TFace4.InitTransform;
var
  I:	Integer;
begin
  for I := 1 to 4 do
    Vertice[I]^.InitTransform;
end;

procedure TFace4.DoneTransform;
var
  I:	Integer;
begin
  for I := 1 to 4 do
    Vertice[I]^.DoneTransform;
end;

{$ifDef Gouraud}
procedure TFace4.ComputeBrightness;
var
  I:		Integer;
begin
  for I := 1 to 4 do with Vertice[I]^ do
  begin
    if Normal.Z > 0 then Inc( Color, Normal.Z div 256 );
    Inc( Divisor );
  end;
end;
{$EndIf}

procedure TFace4.Draw( Buffer: Pointer );
var
  LU,RU,
  LB,RB:	TPoint;
begin
  if Texture <> nil then
  begin
    LU.X := Vertice[1]^.X shr 8 + HRes div 2;
    LU.Y := Vertice[1]^.Y shr 8 + VRes div 2;
    RU.X := Vertice[2]^.X shr 8 + HRes div 2;
    RU.Y := Vertice[2]^.Y shr 8 + VRes div 2;
    LB.X := Vertice[4]^.X shr 8 + HRes div 2;
    LB.Y := Vertice[4]^.Y shr 8 + VRes div 2;
    RB.X := Vertice[3]^.X shr 8 + HRes div 2;
    RB.Y := Vertice[3]^.Y shr 8 + VRes div 2;
    MapRectToPic( LU, RU, RB, LB, Texture, Buffer );
  end;
end;

procedure TFace4.Move( DX, DY, DZ: LongInt );
begin
  Vertice[1]^.Move( DX, DY, DZ );
  Vertice[2]^.Move( DX, DY, DZ );
  Vertice[3]^.Move( DX, DY, DZ );
  Vertice[4]^.Move( DX, DY, DZ );
end;


procedure TFace4.Rotate( CX, CY, CZ: longInt; RX, RY, RZ: Integer );
begin
  Vertice[1]^.Rotate( CX, CY, CZ, RX, RY, RZ );
  Vertice[2]^.Rotate( CX, CY, CZ, RX, RY, RZ );
  Vertice[3]^.Rotate( CX, CY, CZ, RX, RY, RZ );
  Vertice[4]^.Rotate( CX, CY, CZ, RX, RY, RZ );
end;

procedure TFace4.ComputeSortValue;
begin
  SortValue := (Vertice[1]^.Z + Vertice[2]^.Z + Vertice[3]^.Z + Vertice[4]^.Z) div 4;
end;

procedure TFace4.GetScreenExtent( var Bounds: TRect );
var
  I:	Integer;
begin
  Bounds.A.X := 256; Bounds.A.Y := 256;
  Bounds.B.X := 0; Bounds.B.Y := 0;
  for I := 1 to 4 do
  begin
    if Vertice[I]^.X < Bounds.A.X then Bounds.A.X := Vertice[I]^.X;
    if Vertice[I]^.Y < Bounds.A.Y then Bounds.A.Y := Vertice[I]^.Y;
    if Vertice[I]^.X > Bounds.B.X then Bounds.B.X := Vertice[I]^.X;
    if Vertice[I]^.Y > Bounds.B.Y then Bounds.B.Y := Vertice[I]^.Y;
  end;
end;

function TFace4.Visible: Boolean;
begin
  Visible := Normal.Z > 0;
end;

destructor TFace4.Done;
begin
  Vertice[1]^.Free;
  Vertice[2]^.Free;
  Vertice[3]^.Free;
  Vertice[4]^.Free;
end;

var
  I :	Word;
begin
  for I := 0 to CosSteps-1 do
  begin
    CosTable[I] := Round( Cos( 2*Pi*I/CosSteps )*1024 );
    SinTable[I] := Round( Sin( 2*Pi*I/CosSteps )*1024 );
  end;
end.