{$I DEBUG}
unit RStream;

interface

uses Objects;

const
  ResourceFile = 'GLITCH.RES';

type
  PRStream = ^TRStream;
  TRStream = object(TBufStream)
    BasePos,
    EndPos	: LongInt;
{$IFDEF __BPREAL__}
    EMS		: PEMSStream;
{$ELSE}
    EMS		: PMemoryStream;
{$ENDIF}
    constructor Init( FileName: String );
    procedure Seek( Pos: Longint ); virtual;
    function GetSize: Longint; virtual;
    function GetPos: Longint; virtual;
    procedure Read(var Buf; Count: Word); virtual;
    destructor Done; virtual;
  end;
  procedure LoadEMSCopy;
  procedure FreeEMSCopy;
  function GetLine( var F:TStream ): String;

implementation

uses VGAGraph, CRT;

var
  ExeStream: 	TBufStream;
  NFiles:	Integer;
  FilesList:	array[1..80] of record
				  Name: 	String[13];
				  StartOffset,
				  EndOffset:	LongInt;
				end;
{$IFDEF __BPREAL__}
  EMSCopy:	array[1..80] of PEMSStream;
{$ELSE}
  EMSCopy:	array[1..80] of PMemoryStream;
{$ENDIF}

  S: 	String;
  I:    Integer;

function GetLine( var F:TStream ): String;
const
  Spaces = [' ',#9,#10,#13,':',','];
var
  S:    String;
  C:    Char;
begin
  S := '';
  C := #0;
  repeat F.Read( C, 1 ) until not (C in Spaces);
  while (C <> #13) do
  begin
    S := S + C;
    F.Read( C, 1 )
  end;
  while S[ Length(S) ] = ' ' do Dec(S[0]);
  if Pos( 'EOF', S ) <> 0 then
   F.Status := stReadError;
  GetLine := S;
end;

procedure LoadEMSCopy;
var
  S: 	TRStream;
begin
{$IFDEF DEBUG}
  WriteLn( MemF, 'Loading EMS copy : ', NFiles, ' files' );
{$ENDIF}
  for I := 1 to NFiles do
  begin
    S.Init( FilesList[I].Name );
{$IFDEF __BPREAL__}
    New( EMSCopy[I], Init( S.GetSize, S.GetSize ) );
{$ELSE}
    New( EMSCopy[I], Init( S.GetSize, $2000 ) );
{$ENDIF}
    if EMSCopy[I] <> nil then
    begin
      EMSCopy[I]^.CopyFrom(S,S.GetSize);
{$IFDEF DEBUG}
      WriteLn( MemF, FilesList[I].Name, ' copied to EMS (',S.GetSize,')' );
{$ENDIF}
    end
    else
    begin
{$IFDEF DEBUG}
      WriteLn( MemF, FilesList[I].Name, ' not copied to EMS (',S.GetSize,')' );
{$ENDIF}
    end;
    S.Done;
  end;
end;

procedure FreeEMSCopy;
begin
  for I := 1 to NFiles do
  begin
    if EMSCopy[I] <> nil then EMSCopy[I]^.Free;
  end;
end;

constructor TRStream.Init( FileName: String );
begin
{$IFDEF DEBUG}
  WriteLn( MemF, 'Opening ', FIleName );
{$ENDIF}
  inherited Init( FileName, stOpenRead, 4096 );
  EMS := nil;
  if Status <> stOk then
  begin
    for I := 1 to NFiles do
    begin
      if FilesList[I].Name = FileName then
      begin
	Reset;
	if EMSCopy[I] = nil then
	begin
{$IFDEF DEBUG}
	  WriteLn( MemF, 'Found resource ', FIleName );
{$ENDIF}
	  ExeStream.Init( ResourceFile, stOpenRead, 4096 );
	  Handle := ExeStream.Handle;
	  BasePos := FilesList[I].StartOffset;
	  EndPos := FilesList[I].EndOffset;
	  ExeStream.Seek(BasePos);
	  Exit;
	end
	else
	begin
{$IFDEF DEBUG}
	  WriteLn( MemF, 'Found EMS copy ', FIleName );
{$ENDIF}
	  BasePos := 0;
	  EndPos := EMSCopy[I]^.GetSize;
	  EMS := EMSCopy[I];
	  EMS^.Reset;
	  EMS^.Seek(0);
	end;
      end;
    end;
  end
  else
  begin
{$IFDEF DEBUG}
    WriteLn( MemF, 'Found file ', FIleName );
{$ENDIF}
    BasePos := 0;
  end;
end;

procedure TRStream.Seek( Pos: Longint );
begin
  if EMS = nil then
    inherited Seek( Pos+BasePos )
  else
    EMS^.Seek(Pos);
end;

function TRStream.GetPos: Longint;
begin
  if EMS = nil then
    GetPos := inherited GetPos - BasePos
  else
    GetPos := EMS^.GetPos;
end;

function TRStream.GetSize: Longint;
begin
  if EMS = nil then
  begin
    if BasePos > 0 then
      GetSize := EndPos-BasePos
    else
      GetSize := inherited GetSize;
  end
  else
  begin
    GetSize := EMS^.GetSize;
  end;
end;

procedure TRStream.Read(var Buf; Count: Word);
begin
  if EMS = nil then
    inherited Read(Buf,Count)
  else
  begin
    EMS^.Read(Buf,Count);
  end;
end;

destructor TRStream.Done;
begin
  if EMS = nil then
  begin
    if BasePos > 0 then ExeStream.Done;
  end;
  inherited Done;
end;

const
  ListID: String = '#ileList';

begin
  ExeStream.Init( ResourceFile, stOpenRead+$30, 4096 );
{$IFDEF DEBUG}
  if ExeStream.Status <> stOk then
  begin
    WriteLn( MemF, 'Can''t open ', ResourceFile, ' error :', ExeStream.Status );
    ReadKey;
  end;
{$ENDIF}
  ListID[1] := 'F';
  S[0] := #255;
  I := 0;
  repeat
    ExeStream.Read( S[1], 255 );
    if Pos( ListID, S ) > 0 then
    begin
      ExeStream.Reset;
      ExeStream.Seek( LongInt(I)*255+Pos( ListID, S )+7 );
      Break;
    end;
    Inc(I);
  until ExeStream.Status <> stOk;
  if ExeStream.Status = stOk then
  begin
    ExeStream.Read( NFiles, 2 );
    ExeStream.Read( FilesList, 22*NFiles );
  end;
  ExeStream.Done;
  for I := 1 to 80 do
    EMSCopy[I] := nil;
end.