{$I DEBUG}

unit VGAGraph;

{$L SCALE1.OBJ}
{$L SCALE2.OBJ}
{$L SCALE3.OBJ}

interface

uses Objects, Memory, Frames;

type
  PBuffer = ^TBuffer;
  TBuffer = array[ 0..65534 ] of Byte;

  PScreenBuffer = ^TScreenBuffer;
  TScreenBuffer	= array[ 0..199, 0..319 ] of Byte;

  TRGBColor = record R,G,B: Byte; end;
  PRGBPalette = ^TRGBPalette;
  TRGBPalette = array[ 0..255 ] of TRGBColor;

  PImage = ^TImage;
  TImage = object( TObject )
    Size:	TPoint;
    Palette:	PRGBPalette;
    Data:	PBuffer;
    constructor Init( ASize: TPoint; Buffer: Pointer; APalette: PRGBPalette );
    constructor Load( Name: String );
    procedure Show( Origin: TPoint; Buffer: Pointer );
    procedure ShowT( Origin: TPoint; Buffer: Pointer );
    procedure ShowTGConst( Origin: TPoint; Buffer: Pointer; Cons: Byte );
    procedure ShowAddConstT( Origin: TPoint; Buffer: Pointer; Cons: Byte );
    procedure ShowAddT( Origin: TPoint; Buffer: Pointer );
    procedure ShowSubT( Origin: TPoint; Buffer: Pointer );
    procedure ShowXorT( Origin: TPoint; Buffer: Pointer );
    procedure ShowRect( Origin: TPoint; NewSize: TPoint; Buffer: Pointer );
    procedure ShowRectT( Origin: TPoint; NewSize: TPoint; Buffer: Pointer );
    procedure ShowRectAddT( Origin: TPoint; NewSize: TPoint; Buffer: Pointer );
    destructor Done; virtual;
  end;

  PSprite = ^TSprite;
  TSprite = object( TImage )
    Origin,
    Step:	TPoint;
    AddConst:	Byte;
    constructor Init( AOrigin, ASize, AStep: TPoint; Buffer: Pointer );
    procedure Show( Buffer: Pointer );
    procedure ShowT( Buffer: Pointer );
    procedure ShowAddT( Buffer: Pointer );
    procedure ShowSubT( Buffer: Pointer );
    procedure ShowXorT( Buffer: Pointer );
    procedure ShowAddConstT( Buffer: Pointer );
    procedure Move( Steps: Integer );
  end;

  PGraphApplication = ^TGraphApplication;
  TGraphApplication = object( TObject )
    constructor Init;
    procedure Run; virtual;
    destructor Done; virtual;
  end;

var
  VRes,
  HRes:		Integer;
  Draw,
  Screen:	PScreenBuffer;
  CurOffset,
  CurPage:	Word;
  Time: 	LongInt absolute 0:$46C;
{$IFDEF DEBUG}
  MemF:		Text;
{$ENDIF}

  procedure CopyPage( Src, Dest: Pointer );
  procedure ClearPage( Src: Pointer );
  procedure SmoothPage( PFrom, PTo: Pointer );
  procedure MoveSeg( FromPtr, ToPtr: Pointer );
  function AllocPage: Pointer;
  procedure SetPalette( Palette: PRGBPalette );
  procedure GetPalette( Palette: PRGBPalette );
  procedure FadeToPalette( Palette: PRGBPalette; Steps: Integer );
  procedure WaitRetrace;
  function DosAllocMem( Amount: LongInt ): Pointer;
  procedure DosFreeMem( P: Pointer );
  function DosMemAvail: LongInt;
  procedure SetModeQ;
  procedure SetMode13;
  procedure SetModeX360x480;
  procedure SetModeX360x240;
  procedure FlipPageX;
  procedure ClearPageX;
  procedure DrawHLine( Y, X1, X2: Integer; Color: Byte; Buffer: Pointer );
const
  PDrawHLine: Pointer = @DrawHLine;
  procedure Smooth( Src, Dst: PImage );


implementation

uses PCX, CRT
{$IFNDEF __BPREAL__}
,WinAPI
{$ENDIF}
;

procedure SCALE1( OriginX, OriginY, NewSizeX, NewSizeY,
		  SizeX, SizeY: Integer; Data, Buffer: Pointer ); far; external;
procedure SCALE2( OriginX, OriginY, NewSizeX, NewSizeY,
		  SizeX, SizeY: Integer; Data, Buffer: Pointer ); far; external;
procedure SCALE3( OriginX, OriginY, NewSizeX, NewSizeY,
		  SizeX, SizeY: Integer; Data, Buffer: Pointer ); far; external;

constructor TImage.Init(ASize: TPoint; Buffer: Pointer; APalette: PRGBPalette);
begin
  inherited Init;
  Size := ASize;
  Data := DosAllocMem( LongInt(ASize.X)*ASize.Y );
  if APalette <> nil then
  begin
    Palette := MemAlloc( SizeOf(TRGBPalette) );
    Move( APalette, Palette^, SizeOf(TRGBPalette) );
  end
  else
    Palette := nil;
  if Data = nil then Fail;
  if Buffer <> nil then
    Move( Buffer, Data^, Word(ASize.X)*ASize.Y )
  else
  begin
    if LongInt(ASize.X)*ASize.Y < 65535 then
      FillChar( Data^, LongInt(ASize.X)*ASize.Y, 0 )
    else
      FillChar( Data^, 65535, 0 )
  end;
end;

constructor TImage.Load( Name: String );
var
  I:	Integer;
  P:	OldPCXPicture;
begin
  inherited Init;
  P.Init( Name );
{$IFDEF DEBUG}
  WriteLn( MemF, 'Loading ', Name );
{$ENDIF}
  if P.Status <> pcxOK then
  begin
{$IFDEF DEBUG}
    WriteLn( MemF, Name, ' is not a valid PCX' );
{$ENDIF}
    P.Done;
    Fail;
  end;
  Size.X := P.H.XMax - P.H.XMin + 1;
  Size.Y := P.H.YMax - P.H.YMin + 1;

{$IFDEF DEBUG}
  WriteLn( MemF, Name, ' size is ', Size.X, 'x', Size.Y );
{$ENDIF}
  Data := DosAllocMem( LongInt(Size.X) * Size.Y );
  if Data = nil then
  begin
    P.Done;
    Fail;
  end;

  Palette := MemAlloc( SizeOf(TRGBPalette) );
  if Palette <> nil then
  begin
    Move( P.Pal, Palette^, SizeOf(Palette^) );
    for I := 0 to 255 do
    begin
      Palette^[I].R := Palette^[I].R shr 2;
      Palette^[I].G := Palette^[I].G shr 2;
      Palette^[I].B := Palette^[I].B shr 2;
    end;
  end;

  for I := 0 to Size.Y-1 do
    P.ReadLine( Data^[ Word(Size.X)*I ] );

  P.Done;
end;

procedure TImage.Show( Origin: TPoint; Buffer: Pointer );
var
  X,Y:		Integer;
  P:		Pointer;
  AddY,
  AddX,
  SubX,
  SizeX:	Integer;
begin
  X := Size.X; SizeX := X;
  Y := Size.Y;
  if (Origin.Y >= VRes) or (Origin.X >= HRes) then Exit;
  if Origin.Y+Y >= VRes then Y := VRes-Origin.Y;
  if Origin.Y < 0 then AddY := -Origin.Y else AddY := 0;
  if AddY >= Y then Exit;
  Dec( Y, AddY );
  if Origin.X < 0 then AddX := -Origin.X else AddX := 0;
  if Origin.X+X > HRes-1 then SubX := Origin.X+X-HRes else SubX := 0;
  Dec( X, AddX ); Dec( X, SubX );
  if X <= 0 then Exit;
  P := Data;
  asm
    push ds
    les di,Buffer
    mov ax,[Origin.Y]
    test ax,$8000
    je @no_up_clip
    xor ax,ax
@no_up_clip:
    mov	di,HRes
    mul di
    add ax,[Origin.X]
    add ax,AddX
    mov di,ax
    mov ax,SizeX
    mul AddY
    mov	cx,Y
    mov	bx,HRes
    sub	bx,X
    lds	si,P
    add si,ax
@loop_y:
    add si,AddX
    push cx
    mov cx,X
@loop_x:
    lodsb
{
    or al,al
    je @end_loop
}
    mov	es:[di],al
@end_loop:
    inc di
    loop @loop_x
    pop cx
    add di,bx
    add si,SubX
    loop @loop_y
    pop ds
  end;
end;

procedure TImage.ShowT( Origin: TPoint; Buffer: Pointer );
var
  X,Y:		Integer;
  P:		Pointer;
  AddY,
  AddX,
  SubX,
  SizeX:	Integer;
begin
  X := Size.X; SizeX := X;
  Y := Size.Y;
  if (Origin.Y >= VRes) or (Origin.X >= HRes) then Exit;
  if Origin.Y+Y >= VRes then Y := VRes-Origin.Y;
  if Origin.Y < 0 then AddY := -Origin.Y else AddY := 0;
  if AddY >= Y then Exit;
  Dec( Y, AddY );
  if Origin.X < 0 then AddX := -Origin.X else AddX := 0;
  if Origin.X+X > HRes-1 then SubX := Origin.X+X-HRes else SubX := 0;
  Dec( X, AddX ); Dec( X, SubX );
  if X <= 0 then Exit;
  P := Data;
  asm
    push ds
    les di,Buffer
    mov ax,[Origin.Y]
    test ax,$8000
    je @no_up_clip
    xor ax,ax
@no_up_clip:
    mov	di,HRes
    mul di
    add ax,[Origin.X]
    add ax,AddX
    mov di,ax
    mov	cx,Y
    mov	bx,HRes
    sub	bx,X
    lds	si,P
    mov ax,SizeX
    mul AddY
    add si,ax
@loop_y:
    add si,AddX
    push cx
    mov cx,X
@loop_x:
    lodsb
    or al,al
    je @end_loop
    mov	es:[di],al
@end_loop:
    inc di
    loop @loop_x
    pop cx
    add di,bx
    add si,SubX
    loop @loop_y
    pop ds
  end;
end;

procedure TImage.ShowTGConst( Origin: TPoint; Buffer: Pointer; Cons: Byte );
var
  X,Y:		Integer;
  P:		Pointer;
  AddY,
  AddX,
  SubX,
  SizeX:	Integer;
begin
  X := Size.X; SizeX := X;
  Y := Size.Y;
  if (Origin.Y >= VRes) or (Origin.X >= HRes) then Exit;
  if Origin.Y+Y >= VRes then Y := VRes-Origin.Y;
  if Origin.Y < 0 then AddY := -Origin.Y else AddY := 0;
  if AddY >= Y then Exit;
  Dec( Y, AddY );
  if Origin.X < 0 then AddX := -Origin.X else AddX := 0;
  if Origin.X+X > HRes-1 then SubX := Origin.X+X-HRes else SubX := 0;
  Dec( X, AddX ); Dec( X, SubX );
  if X <= 0 then Exit;
  P := Data;
  asm
    mov dl,Cons
    push ds
    les di,Buffer
    mov ax,[Origin.Y]
    test ax,$8000
    je @no_up_clip
    xor ax,ax
@no_up_clip:
    mov	di,HRes
    mul di
    add ax,[Origin.X]
    add ax,AddX
    mov di,ax
    mov	cx,Y
    mov	bx,HRes
    sub	bx,X
    lds	si,P
    mov ax,SizeX
    mul AddY
    add si,ax
@loop_y:
    add si,AddX
    push cx
    mov cx,X
@loop_x:
    lodsb
    cmp al,dl
    jbe @end_loop
    mov	es:[di],al
@end_loop:
    inc di
    loop @loop_x
    pop cx
    add di,bx
    add si,SubX
    loop @loop_y
    pop ds
  end;
end;

procedure TImage.ShowAddConstT( Origin: TPoint; Buffer: Pointer; Cons: Byte );
var
  X,Y:		Integer;
  P:		Pointer;
  AddY,
  AddX,
  SubX,
  SizeX:	Integer;
begin
  X := Size.X; SizeX := X;
  Y := Size.Y;
  if (Origin.Y >= VRes) or (Origin.X >= HRes) then Exit;
  if Origin.Y+Y >= VRes then Y := VRes-Origin.Y;
  if Origin.Y < 0 then AddY := -Origin.Y else AddY := 0;
  if AddY >= Y then Exit;
  Dec( Y, AddY );
  if Origin.X < 0 then AddX := -Origin.X else AddX := 0;
  if Origin.X+X > HRes-1 then SubX := Origin.X+X-HRes else SubX := 0;
  Dec( X, AddX ); Dec( X, SubX );
  if X <= 0 then Exit;
  P := Data;
  asm
    push ds
    les di,Buffer
    mov ax,[Origin.Y]
    test ax,$8000
    je @no_up_clip
    xor ax,ax
@no_up_clip:
    mov	di,HRes
    mul di
    add ax,[Origin.X]
    add ax,AddX
    mov di,ax
    mov	cx,Y
    mov	bx,HRes
    sub	bx,X
    lds	si,P
    mov ax,SizeX
    mul AddY
    add si,ax
@loop_y:
    add si,AddX
    push cx
    mov cx,X
@loop_x:
    lodsb
    or al,al
    je @end_loop
    add al,Cons
    mov	es:[di],al
@end_loop:
    inc di
    loop @loop_x
    pop cx
    add di,bx
    add si,SubX
    loop @loop_y
    pop ds
  end;
end;

procedure TImage.ShowAddT( Origin: TPoint; Buffer: Pointer );
var
  X,Y:		Integer;
  P:		Pointer;
  AddY,
  AddX,
  SubX,
  SizeX:	Integer;
begin
  X := Size.X; SizeX := X;
  Y := Size.Y;
  if (Origin.Y >= VRes) or (Origin.X >= HRes) then Exit;
  if Origin.Y+Y >= VRes then Y := VRes-Origin.Y;
  if Origin.Y < 0 then AddY := -Origin.Y else AddY := 0;
  if AddY >= Y then Exit;
  Dec( Y, AddY );
  if Origin.X < 0 then AddX := -Origin.X else AddX := 0;
  if Origin.X+X > HRes-1 then SubX := Origin.X+X-HRes else SubX := 0;
  Dec( X, AddX ); Dec( X, SubX );
  if X <= 0 then Exit;
  P := Data;
  asm
    push ds
    les di,Buffer
    mov ax,[Origin.Y]
    test ax,$8000
    je @no_up_clip
    xor ax,ax
@no_up_clip:
    mov	di,HRes
    mul di
    add ax,[Origin.X]
    add ax,AddX
    mov di,ax
    mov	cx,Y
    mov	bx,HRes
    sub	bx,X
    lds	si,P
    mov ax,SizeX
    mul AddY
    add si,ax
@loop_y:
    add si,AddX
    push cx
    mov cx,X
@loop_x:
    lodsb
    or al,al
    je @end_loop
    add	es:[di],al
@end_loop:
    inc di
    loop @loop_x
    pop cx
    add di,bx
    add si,SubX
    loop @loop_y
    pop ds
  end;
end;

procedure TImage.ShowSubT( Origin: TPoint; Buffer: Pointer );
var
  X,Y:		Integer;
  P:		Pointer;
  AddY,
  AddX,
  SubX,
  SizeX:	Integer;
begin
  X := Size.X; SizeX := X;
  Y := Size.Y;
  if (Origin.Y >= VRes) or (Origin.X >= HRes) then Exit;
  if Origin.Y+Y >= VRes then Y := VRes-Origin.Y;
  if Origin.Y < 0 then AddY := -Origin.Y else AddY := 0;
  if AddY >= Y then Exit;
  Dec( Y, AddY );
  if Origin.X < 0 then AddX := -Origin.X else AddX := 0;
  if Origin.X+X > HRes-1 then SubX := Origin.X+X-HRes else SubX := 0;
  Dec( X, AddX ); Dec( X, SubX );
  if X <= 0 then Exit;
  P := Data;
  asm
    push ds
    les di,Buffer
    mov ax,[Origin.Y]
    test ax,$8000
    je @no_up_clip
    xor ax,ax
@no_up_clip:
    mov	di,HRes
    mul di
    add ax,[Origin.X]
    add ax,AddX
    mov di,ax
    mov	cx,Y
    mov	bx,HRes
    sub	bx,X
    lds	si,P
    mov ax,SizeX
    mul AddY
    add si,ax
@loop_y:
    add si,AddX
    push cx
    mov cx,X
@loop_x:
    lodsb
    or al,al
    je @end_loop
    sub	es:[di],al
@end_loop:
    inc di
    loop @loop_x
    pop cx
    add di,bx
    add si,SubX
    loop @loop_y
    pop ds
  end;
end;


procedure TImage.ShowXorT( Origin: TPoint; Buffer: Pointer );
var
  X,Y:		Integer;
  P:		Pointer;
  AddY,
  AddX,
  SubX,
  SizeX:	Integer;
begin
  X := Size.X; SizeX := X;
  Y := Size.Y;
  if (Origin.Y >= VRes) or (Origin.X >= HRes) then Exit;
  if Origin.Y+Y >= VRes then Y := VRes-Origin.Y;
  if Origin.Y < 0 then AddY := -Origin.Y else AddY := 0;
  if AddY >= Y then Exit;
  Dec( Y, AddY );
  if Origin.X < 0 then AddX := -Origin.X else AddX := 0;
  if Origin.X+X > HRes-1 then SubX := Origin.X+X-HRes else SubX := 0;
  Dec( X, AddX ); Dec( X, SubX );
  if X <= 0 then Exit;
  P := Data;
  asm
    push ds
    les di,Buffer
    mov ax,[Origin.Y]
    test ax,$8000
    je @no_up_clip
    xor ax,ax
@no_up_clip:
    mov	di,HRes
    mul di
    add ax,[Origin.X]
    add ax,AddX
    mov di,ax
    mov	cx,Y
    mov	bx,HRes
    sub	bx,X
    lds	si,P
    mov ax,SizeX
    mul AddY
    add si,ax
@loop_y:
    add si,AddX
    push cx
    mov cx,X
@loop_x:
    lodsb
    or al,al
    je @end_loop
    xor	es:[di],al
@end_loop:
    inc di
    loop @loop_x
    pop cx
    add di,bx
    add si,SubX
    loop @loop_y
    pop ds
  end;
end;

procedure TImage.ShowRectT( Origin: TPoint; NewSize: TPoint; Buffer: Pointer );
begin
 SCALE1( Origin.X, Origin.Y, NewSize.X, NewSize.Y, Size.X, Size.Y, Data, Buffer );
end;

procedure TImage.ShowRect( Origin: TPoint; NewSize: TPoint; Buffer: Pointer );
begin
 SCALE2( Origin.X, Origin.Y, NewSize.X, NewSize.Y, Size.X, Size.Y, Data, Buffer );
end;

procedure TImage.ShowRectAddT( Origin: TPoint; NewSize: TPoint; Buffer: Pointer );
begin
 SCALE3( Origin.X, Origin.Y, NewSize.X, NewSize.Y, Size.X, Size.Y, Data, Buffer );
end;

destructor TImage.Done;
begin
  DosFreeMem( Data );
  if Palette <> nil then
    FreeMem( Palette, SizeOf(TRGBPalette) );
  inherited Done;
end;

constructor TSprite.Init( AOrigin, ASize, AStep: TPoint; Buffer: Pointer );
begin
  inherited Init( ASize, Buffer, nil );
  Origin := AOrigin;
  Step := AStep;
end;

procedure TSprite.Show( Buffer: Pointer );
begin
  inherited Show( Origin, Buffer );
end;

procedure TSprite.ShowT( Buffer: Pointer );
begin
  inherited ShowT( Origin, Buffer );
end;

procedure TSprite.ShowAddT( Buffer: Pointer );
begin
  inherited ShowAddT( Origin, Buffer );
end;

procedure TSprite.ShowSubT( Buffer: Pointer );
begin
  inherited ShowSubT( Origin, Buffer );
end;

procedure TSprite.ShowXorT( Buffer: Pointer );
begin
  inherited ShowXorT( Origin, Buffer );
end;

procedure TSprite.ShowAddConstT( Buffer: Pointer );
begin
  inherited ShowAddConstT( Origin, Buffer, AddConst );
end;

procedure TSprite.Move( Steps: Integer );
begin
  Inc( Origin.X, Step.X*Steps );
  Inc( Origin.Y, Step.Y*Steps );
end;

constructor TGraphApplication.Init;
begin
{
  asm
    mov	ax,13h
    int	10h
  end;
}
end;

procedure TGraphApplication.Run;
begin
  Abstract;
end;

destructor TGraphApplication.Done;
begin
  asm
    mov	ax,3h
    int	10h
  end;
end;

procedure CopyPage( Src, Dest: Pointer );
begin
  if KeyPressed then Frames.Abort;
  if Dest = Screen then WaitRetrace;
  asm
    push ds
    xor dx,dx
    mov ax,VRes
    mul HRes
    shr dx,1
    rcr ax,1
    shr dx,1
    rcr ax,1
    mov cx,ax
    lds si,Src
    les di,Dest
    cld
    db $66; rep movsw
    pop ds
  end;
end;

procedure ClearPage( Src: Pointer );
begin
  asm
    xor dx,dx
    mov ax,VRes
    mul HRes
    shr dx,1
    rcr ax,1
    shr dx,1
    rcr ax,1
    mov cx,ax
    les di,Src
    db $66; xor ax,ax
    cld
    db $66; rep stosw
  end;
end;

procedure SmoothPage( PFrom, PTo: Pointer );
begin
  asm
    push ds
    les di,PTo
    lds si,PFrom
    xor ax,ax
    mov cx,160
    rep stosw
    add si,320
    mov cx,64000-320*2
@loop:
      mov al,[si+320]		{ - - - }
      xor ah,ah                 { - - - }
				{ - * - }

      add al,[si-320]		{ - * - }
      adc ah,0                  { - - - }
				{ - - - }

      add al,[si-1]             { - - - }
      adc ah,0                  { * - - }
				{ - - - }

      add al,[si+1]             { - - - }
      adc ah,0                  { - - * }
				{ - - - }
(*
      add al,[si-320-1]         { * - - }
      adc ah,0                  { - - - }
				{ - - - }

      add al,[si+320+1]         { - - - }
      adc ah,0                  { - - - }
				{ - - * }

      add al,[si-320+1]         { - - * }
      adc ah,0                  { - - - }
				{ - - - }

      add al,[si+320-1]         { - - - }
      adc ah,0                  { - - - }
				{ * - - }
      shr ax,3
*)
      shr ax,2
      stosb
      inc si
    loop @loop
    xor ax,ax
    mov cx,160
    rep stosw
    pop ds
  end;
{
  Tmp := PFrom;
  PFrom := PTo;
  PTo := Tmp;
}
end;

procedure SetPalette( Palette: PRGBPalette );
begin
  WaitRetrace;
  asm
    push ds
    lds si, Palette
    xor al,al
    mov cx,768
    mov dx,$3C8
    out dx,al
    inc dx
    rep outsb
    pop ds
  end;
end;

procedure GetPalette( Palette: PRGBPalette );
begin
  asm
    les di, Palette
    xor al,al
    mov cx,768
    mov dx,$3C7
    out dx,al
    add dx,2
    rep insb
  end;
end;

procedure WaitRetrace;
begin
  repeat until (Port[$3DA] and $08) = 0;
  repeat until (Port[$3DA] and $08) <> 0;
end;

{$IFDEF __BPREAL__}
function DosAllocMem( Amount: LongInt ): Pointer;
var
  Tmp:	Pointer;
begin
  Tmp := nil;
  asm
    mov ah,48h
    mov dx,word ptr Amount+2
    mov bx,word ptr Amount
    add bx,15
    adc dx,0
    shr dx,1
    rcr bx,1
    shr dx,1
    rcr bx,1
    shr dx,1
    rcr bx,1
    shr dx,1
    rcr bx,1
    int 21h
    jnc @mem_ok
    xor ax,ax
@mem_ok:
    mov word ptr Tmp+2,ax
  end;
{$IFDEF DEBUG}
  WriteLn( MemF, 'DosAllocMem( ', Amount,' ) : ',#9#9,DosMemAvail, #9#9, MemAvail );
{$ENDIF}
  DosAllocMem := Tmp;
end;
{$ELSE}
function DosAllocMem( Amount: LongInt ): Pointer;
var
  Tmp:	Pointer;
  H:	THandle;
begin
  DosAllocMem := GlobalAllocPtr( GMEM_ZEROINIT or GMEM_FIXED, Amount );
{$IFDEF DEBUG}
  WriteLn( MemF, 'DosAllocMem( ', Amount,' ) : ',#9#9,DosMemAvail, #9#9, MemAvail );
{$ENDIF}
end;
{$ENDIF}

{$IFDEF __BPREAL__}
procedure DosFreeMem( P: Pointer );
begin
  if P = nil then Exit;
asm
  mov ax,word ptr P+2
  mov es,ax
  mov ah,49h
  int 21h
end;
{$IFDEF DEBUG}
  WriteLn( MemF, 'DosFreeMem( ',LongInt(P) shr 16,' ) : ',DosMemAvail );
{$ENDIF}
end;
{$ELSE}
procedure DosFreeMem( P: Pointer );
begin
  if P = nil then Exit;
  GlobalFreePtr(P);
{$IFDEF DEBUG}
  WriteLn( MemF, 'DosFreeMem( ',LongInt(P) shr 16,' ) : ',DosMemAvail );
{$ENDIF}
end;
{$ENDIF}

{$IFDEF __BPREAL__}
function DosMemAvail: LongInt; assembler;
asm
    mov ah,48h
    mov bx,$FFFF
    int 21h
    mov ax,bx
    shl ax,4
    mov dx,bx
    shr dx,12
end;
{$ELSE}
function DosMemAvail: LongInt;
begin
  DosMemAvail := GetFreeSpace(0);
end;
{$ENDIF}

type
  twrec=record reg:word; func,data:byte; end;
  twarr=array[0..22] of twrec;

const
  CRTModeQ:twarr=(
    (reg:$03d4; func:$00; data:$5f), { hor. total }
    (reg:$03d4; func:$01; data:$3f), { hor. display enable end }
    (reg:$03d4; func:$02; data:$40), { blank start }
    (reg:$03d4; func:$03; data:$82), { blank end }
    (reg:$03d4; func:$04; data:$4e), { retrace start }
    (reg:$03d4; func:$05; data:$9a), { retrace end }
    (reg:$03d4; func:$06; data:$23), { vertical total }
    (reg:$03d4; func:$07; data:$b2), { overflow register }
    (reg:$03d4; func:$08; data:$00), { preset row scan }
    (reg:$03d4; func:$09; data:$61), { max scan line/char heigth }
    (reg:$03d4; func:$10; data:$0a), { ver. retrace start }
    (reg:$03d4; func:$11; data:$ac), { ver. retrace end }
    (reg:$03d4; func:$12; data:$ff), { ver. display enable end }
    (reg:$03d4; func:$13; data:$20), { offset/logical width }
    (reg:$03d4; func:$14; data:$40), { underlinde location }
    (reg:$03d4; func:$15; data:$07), { ver. blank start }
    (reg:$03d4; func:$16; data:$17), { ver. blank end }
    (reg:$03d4; func:$17; data:$a3), { mode control }
    (reg:$03c4; func:$01; data:$01), { clock mode register }
    (reg:$03c4; func:$04; data:$0e), { memory mode register }
    (reg:$03ce; func:$05; data:$40), { mode register }
    (reg:$03ce; func:$06; data:$05), { misc. register }
    (reg:$03c0; func:$10; data:$41)  { mode control }
  );

procedure openregs; assembler; asm
  mov dx,03d4h; mov al,11h; out dx,al; inc dx; in al,dx; and al,7fh
  mov ah,al; mov al,11h; dec dx; out dx,ax; end;

procedure closeregs; assembler; asm
  mov dx,03d4h; mov al,11h; out dx,al; inc dx; in al,dx; or al,80h
  mov ah,al; mov al,11h; dec dx; out dx,ax; end;

procedure SetModeQ;
var i:byte;
begin
  SetMode13;
  openregs;
  for i:=0 to 22 do
    with CRTModeQ[i] do begin
      if reg<>$03c0 then port[reg]:=func else port[reg]:=func+32;
      port[reg+1]:=data;
    end;
  closeregs;
  VRes := 256; HRes := 256;
end;

procedure FadeToPalette( Palette: PRGBPalette; Steps: Integer );
var
  I, J:	Integer;
  From,
  Pal:	TRGBPalette;
begin
  GetPalette( @From );
  Pal := Palette^;
  for I := 1 to Steps do
  begin
    for J := 0 to 255 do
    begin
      Pal[J].R := From[J].R + (Palette^[J].R-From[J].R)*I div Steps;
      Pal[J].G := From[J].G + (Palette^[J].G-From[J].G)*I div Steps;
      Pal[J].B := From[J].B + (Palette^[J].B-From[J].B)*I div Steps;
    end;
    WaitNextFrame;
    SetPalette( @Pal );
  end;
end;

procedure SetMode13;
var
  Pal:	TRGBPalette;
begin
  FillChar( Pal, SizeOf(Pal), 0 );
  SetPalette(@Pal);
  VRes := 200; HRes := 320;
  asm
    mov ax,13h
    int 10h
  end;
end;

function AllocPage: Pointer;
begin
  AllocPage := DosAllocMem( LongInt(HRes)*VRes );
end;

const
  CRTTableLength=17;

type
  TCRTTable = array [1..CRTTableLength] of record func, data: Byte; end;

const
  CRTTable360x480:TCRTTable=(
    (func:$00; data:$6b), { hor. total }
    (func:$01; data:$59), { hor. display enable end }
    (func:$02; data:$5a), { blank start }
    (func:$03; data:$8e), { blank end }
    (func:$04; data:$5e), { retrace start }
    (func:$05; data:$8a), { retrace end }
    (func:$06; data:$0d), { vertical total }
    (func:$07; data:$3e), { overflow register }
(*    (func:$08; data:$00), { preset row scan }*)
    (func:$09; data:$40), { max scan line/char heigth }
    (func:$10; data:$ea), { ver. retrace start }
    (func:$11; data:$8c), { ver. retrace end }
    (func:$12; data:$df), { ver. display enable end }
    (func:$13; data:$2d), { offset/logical width }
    (func:$14; data:$00), { underlinde location }
    (func:$15; data:$e7), { ver. blank start }
    (func:$16; data:$06), { ver. blank end }
    (func:$17; data:$e3) { mode control }
  );

const
  CRTTable360x240:TCRTTable=(
    (func:$00; data:$6b), { hor. total }
    (func:$01; data:$59), { hor. display enable end }
    (func:$02; data:$5a), { blank start }
    (func:$03; data:$8e), { blank end }
    (func:$04; data:$5e), { retrace start }
    (func:$05; data:$8a), { retrace end }
    (func:$06; data:$0d), { vertical total }
    (func:$07; data:$3e), { overflow register }
(*    (func:$08; data:$00), { preset row scan }*)
    (func:$09; data:$41), { max scan line/char heigth }
    (func:$10; data:$ea), { ver. retrace start }
    (func:$11; data:$8c), { ver. retrace end }
    (func:$12; data:$df), { ver. display enable end }
    (func:$13; data:$2d), { offset/logical width }
    (func:$14; data:$00), { underlinde location }
    (func:$15; data:$e7), { ver. blank start }
    (func:$16; data:$06), { ver. blank end }
    (func:$17; data:$e3) { mode control }
  );

procedure SetModeX( var CRTTable: TCRTTable );
var
  i:		byte;
  Black:	TRGBPalette;
begin

  SetMode13;

  FillChar( Black, SizeOf(Black), 0 );
  SetPalette( @Black );
  PortW[$3C4] := $604;
  PortW[$3C4] := $100;
  Port[$3C2] := $E7;
  PortW[$3C4] := $300;
  Port[$3D4] := $11;
  Port[$3D5] := Port[$3D5] and $7F;
  for i:=1 to CRTTableLength do
    with CRTTable[i] do
    begin
      PortW[$3D4] := func;
      PortW[$3D4+1] := data;
    end;
  PortW[$3C4] := $F02;
  ClearPage( Screen );
end;

procedure SetModeX360x480;
begin
  SetModeX( CRTTable360x480 );
  VRes := 480; HRes := 360;
end;

procedure SetModeX360x240;
begin
  SetModeX( CRTTable360x240 );
  VRes := 240; HRes := 360;
end;

procedure FlipPageX;
begin
{  MoveSeg( Screen, Ptr( SegA000, 0 ) );}
  asm
      mov bx,CurOffset
      mov dx,3d4h
      mov al,0ch
      mov ah,bh
      out dx,ax
      mov dx,3d4h
      mov al,0dh
      mov ah,bl
      out dx,ax

      mov dx,3dah
 @@1: in al,dx
      test al,00001000b
      jz @@1
 @@2: in   al,dx
      test al,00001000b
      jnz  @@2

  end;
  Inc(CurPage);
  CurPage := CurPage mod 3;
  CurOffset := CurPage*HRes div 4 *VRes;
  Screen := Ptr( SegA000, CurOffset );
end;

procedure ClearPageX;
begin
{  ClearSeg( Screen );}
  asm
    mov     dx,$3C4
    mov     ax,0f02h
    out     dx,ax
    les     di,Screen
    mov	    cx,HRes
    shr	    cx,2
    mov     ax,VRes
    xor dx,dx
    mul cx
    mov cx,ax
    xor     ax,ax
    rep     stosw
  end;
end;

procedure DrawHLine( Y, X1, X2: Integer; Color: Byte; Buffer: Pointer );
begin
  if (Y < VRes) or (Y >= 0) then
  begin
    asm
      mov ax,X1
      mov cx,X2
      cmp ax,cx
      jle @no_swap
      xchg ax,cx
@no_swap:
      cmp ax,0
      jge @no_l
      xor ax,ax
@no_l:
      cmp ax,HRes
      jge @exit
      cmp cx,HRes
      jl @no_r
      mov cx,HRes
      dec cx
@no_r:
      cmp cx,0
      jl @exit
      sub cx,ax
      inc cx
      les di,Buffer
      mov di,ax
      mov ax,HRes
      mul Y
      add di,ax
      mov al,Color
      rep stosb
@exit:
    end;
  end;
end;

procedure Smooth( Src, Dst: PImage );
var
  P1,
  P2:	Pointer;
  H,
  V:	Integer;
begin
  P1 := Src^.Data;
  P2 := Dst^.Data;
  H := Src^.Size.X-2; V := Src^.Size.Y-2;
  asm
    push ds
    mov bx,H
    les di,P2
    lds di,P1
    mov cx,bx
    add cx,2
    xor ax,ax
    rep stosb
    mov cx,V
@loopY:
    mov es:[di],ah
    inc di
    push cx
    mov cx,bx
@loopX:
    mov al,[di+1]
    add al,[di-1]
    add al,[di+bx]
    neg bx
    add al,[di+bx]
    neg bx
    shr ax,2
    mov es:[di],al
    inc di
    dec cx
    jnz @loopX

    mov es:[di],ah
    inc di
    pop cx
    dec cx
    jnz @loopY

    mov cx,bx
    add cx,2
    xor al,al
    rep stosb
    pop ds
  end;
end;

procedure MoveSeg( FromPtr, ToPtr: Pointer );
begin
  WaitRetrace;
  asm
    push ds
    lds si,FromPtr
    les di,ToPtr
    mov cx,16384*2
    rep movsw
    pop ds
  end;
end;

begin
  Screen := Ptr( SegA000, 0 );
end.